require 'mixlib/config'
require 'veil'

module Manage
  # This module if for reading the config from /etc/chef-manage/manage.rb.
  #
  # A line in this config file looks like:
  #
  #     support_email_address 'test@test.com'
  #     webapp.port 4567
  #
  # Anything that is configurable by users needs to be defined here with
  # config_contexts and defaults or configurables.
  #
  # The attributes in here are the ones that are configurable in the config
  # file. A full list is in attributes/default.rb
  module Config
    extend Mixlib::Config

    config_strict_mode true

    ## Events
    #
    # The events service manages updating the UI for long-running actions.
    config_context :events do
      # The port on which the events service runs.
      configurable :port
    end

    ## Google Analytics
    #
    # Settings for configuring use of Google Analytics
    config_context :google_analytics do
      # The Google Analytics Tracking ID
      configurable :account
    end

    ## Logging
    #
    # Lets you set log levels
    config_context :logging do
      # The log level used for requests to the Chef server from the web app.
      default :chef_log_level, 'info'

      # The log level for Rails and Sidekiq services.
      default :log_level, 'info'
    end

    ## Marketo
    #
    # Used by Chef Software for marketing purposes.
    config_context :marketo do
      configurable :enabled
      configurable :campaign_id
      configurable :client_id
      configurable :client_secret
      configurable :host
    end

    # Used by Chef Software for marketing purposes.
    config_context :eloqua do
      configurable :enabled
      configurable :form_id
      configurable :host
      configurable :username
      configurable :password
    end

    ## Platform
    #
    # The platform attributes specify information about the Chef server that manage
    # uses to interact with it.
    config_context :platform do
      # The name of the privileged user that can manage requests to the Chef server.
      default :user, 'pivotal'

      # NOTE: The url, ldap.enabled, and ldap.adjective attributes are defined
      # dynamically in the recipe because they use the node's private_chef attributes
      # from EC which are not available here.
    end

    # Set the ssl_verify_mode for Chef Manage communicating with the Chef Infra Server API.
    # The default is :verify_none, the only other option is :verify_peer
    # We default to :verify_none as this is the first version of Manage that contains a version of the Infra Client that has this functionality.
    # This is release contains a major jump from Infra Client 11 to Infra Client 14.
    # If users are using self-signed certificates on the Infra Server, as is the default, and we defaulted to :verify_peer,
    # Manage would fail to authenticate and return an HTTP 500 error to the user
    configurable :ssl_verify_mode

    # Configure trusted_certs_dir path for Chef Infra Server, by default is ""
    # this can be configured using /etc/chef-manage/manage.rb
    configurable :trusted_certs_dir

    # Mtls Settings for manage.
    # These default to "" and can be set via /etc/chef/chef-manage.rb
    # Path to the client key and cert.
    configurable :ssl_client_cert
    configurable :ssl_client_key

    # Optional mtls Settings for manage. These default to "" and can be set via /etc/chef/chef-manage.rb
    configurable :ssl_ca_file

    ## SAML
    #
    # Configure SAML auth via Delivery IDP.
    config_context :saml do
      # Whether SAML auth is enabled
      default :enabled, false

      # The client id
      configurable :client_id

      # The client secret
      configurable :client_secret

      # The IDP endpoint of the Delivery server
      default :issuer_url, 'http://delivery/api/v0'

      # Whether the client should verify the TLS cert
      default :verify_tls, true

      # The adjective used in the Manage UI
      default :adjective, 'SAML'
    end

    ## ReCAPTCHA
    #
    # Configure reCAPTCHA for sign-up page
    config_context :recaptcha do
      default :enabled, false
      default :fail_open, true
      configurable :site_key
      configurable :secret_key
      default :verify_url, 'https://www.google.com/recaptcha/api/siteverify'
    end

    ## Redis
    #
    # Redis is used for background worker queues.
    config_context :redis do
      # The host running redis.
      default :host, 'localhost'

      # The port for redis.
      default :port, 11002

      # The redis URL, in the form of redis://HOST:PORT/DB_NUM
      default(:url) { "redis://#{host}:#{port}/0" }
    end

    ## Sentry
    #
    # Configuration for Sentry error tracking service (https://getsentry.com/welcome/)
    config_context :sentry do
      # Private data source name
      configurable :dsn

      # Public data source name
      configurable :public_dsn
    end

    ## Services
    #
    # These control whether or not the individual services needed to run the
    # application are enabled.
    #
    # Because these currently have dashes in them, you have to use the square-
    # bracket syntax if using them in the config file:
    #
    #     services.events.enable = false
    #
    config_context :services do
      config_context :events do
        configurable :enable
      end
      config_context :web do
        default :enable, true
      end
      config_context :worker do
        default :enable, true
      end
    end

    ## Webapp
    #
    # These attributes control the Unicorn application server that runs the webapp.
    config_context :webapp do
      # The internal port on which the webapp runs.
      default :port, 9462

      # The host and port on which the service listens.
      default(:listen) { "127.0.0.1:#{port}" }

      # These tune Unicorn's TCP listening settings.
      # See http://unicorn.bogomips.org/Unicorn/Configurator.html#method-i-listen
      default :backlog, 1024
      default :tcp_nodelay, true

      # The number of worker processes to run on the machine.
      # See http://unicorn.bogomips.org/Unicorn/Configurator.html#method-i-worker_processes
      #
      # This is defined dynamically in the cookbook's attributes as the number
      # of CPUs on the machine, but can be overridden here.
      configurable :worker_processes

      # The timeout (in seconds) for requests.
      # See http://unicorn.bogomips.org/Unicorn/Configurator.html#method-i-timeout
      default :worker_timeout, 3600

      # The min/max number of requests a Unicorn worker can handle before
      # the unicorn-worker-killer can kill and spawn a new worker
      default :min_request, 3072
      default :max_request, 4096

      # The min/max amount of RAM (in MB) a Unicorn worker can consume before
      # the unicorn-worker-killer can kill and spawn a new worker
      default :min_memory, 120
      default :max_memory, 150

      # Whether we rely on the chef-server do acl checks on search results
      configurable :strict_search_result_acls
    end

    # This is what ends up in Rails' settings.yml
    config_context :secrets do
      default :provider, 'chef-secrets-env'
      default :var_name, 'CHEF_SECRETS_DATA'
    end

    ## Other
    #
    # These are top-level attributes used in the configuration.

    # Amount of time, in milliseconds, after which search requests in the web
    # client will time out and cause an error. Backbone.Paginator's default is
    # 25000 (25s), but the Manage web app and the default here are 120000.
    default :backbone_paginator_timeout, 120000

    # URL of the site used for Chef documentation
    default :chef_documentation_url, 'https://docs.chef.io'

    # Whether or not to turn on DemandBase
    configurable :demandbase

    # Disable sign up and allow only sign in.
    default :disable_sign_up, false

    # The from address that is used for password reset and invitation emails.
    default :email_from_address,
            'Chef Notifications <notifications@chef.io>'

    # The fully qualified domain name for the management console. By default it
    # runs on the same FQDN as the API (lb.api_fqdn in the Chef server
    # configuration.)
    #
    # This doesn't really do anything in Chef server 11, since a webui will
    # always be running on the API FQDN anyway.
    #
    # In Chef server 11, If you want manage to work at the FQDN you specify here
    # but not at the API FQDN, you'll need to set:
    #
    #     allowed_webui_subnets ["127.0.0.1/32"]
    #
    # in /etc/opscode/chef-server.rb.
    configurable :fqdn

    # Addons must be added to the Chef server load balancer in a specific order.
    # This attribute designates the prefix used in this addon to determine that
    # ordering.
    default :nginx_addon_prefix, 30

    # Whether or not users can create new organizations from Chef management console
    default :org_creation_enabled, true

    # The port that the external load balancer uses to access the app. Defaults
    # to the same port as the API (Chef server's nginx.ssl_port)
    configurable :public_port

    # The timeout that runit (the process supervisor) waits for services to respond.
    # See SVWAIT in the environment section on http://smarden.org/runit/sv.8.html
    default :runit_timeout, 30

    # The secret token used by Rails to prevent session tampering.
    # See http://guides.rubyonrails.org/security.html#session-storage
    configurable :secret_token
    configurable :secret_key_base

    # Time (in minutes) of inactivity until a browser session times out. If not set
    # (or set to 0), sessions will not have a timed expiration and expire when the
    # browser session ends (or the absolute session timeout is hit, see below),
    # usually when the browser quits or the computer restarts.
    default :session_timeout, 1440

    # Time (in minutes) until a browser session is reset, regardless of activity. If
    # not set (or set to 0), sessions will not have a timed expiration and will
    # never expire except due to inactivity (see session_timeout above).
    default :session_timeout_absolute, 10080

    # Maximum number of logins failures until a user account is locked and the user
    # must use the password reset feature to reset their password and unlock their
    # account.  Unsetting this or setting this to zero or false will disable this
    # feature.
    default :max_login_failures, 7

    # A message to display on the sign in page. If present, the message will
    # show up in a box on the sign in page. Any valid HTML is allowed here.
    # Default is nil, which means no message will show.
    configurable :sign_in_page_message

    # The email address that the feedback forms asks you to use for sending support
    # requests.
    default :support_email_address, 'support@chef.io'

    # The URL of the site used for links to support
    default :support_site_url, 'https://chef.io/support'

    # The URL of the site used to file support tickets
    default :support_tickets_url, 'https://getchef.zendesk.com/hc/en-us/requests/new'

    # Chef Marketplace configuration options
    config_context :marketplace do
      config_context :sign_up do
        # Enable the marketplace signup flow
        default :enabled, false

        # How long the marketplace signup page is available after reconfigure
        default :timeout, 3600
      end

      config_context :sentinel do
        # The sentinel file location
        default :file, '/var/opt/chef-manage/marketplace_sentinel'
      end
    end

    ## Undocumented features
    #
    # Everything below here is either only used by Chef Software, or
    # undocumented for other purposes.

    # The thing we call Chef
    default :chef_title, 'Chef Infra'

    # Whether or not to show the upgrade banner
    default :disable_upgrade_banner, false

    config_context :user do
      default :username, 'opscode'
      default :shell, '/bin/sh'
      default :home, '/opt/opscode/embedded'
    end

    # Whether airgapped or not 
    default :air_gapped, false
    
    # Configure GTM
    config_context :gtm do
      default :enabled, false
      configurable :id
    end

    # Configure onetrust
    config_context :onetrust do
      default :enabled, false
      configurable :id
    end

    class << self
      def save(_opts)
        [
          {name: "secret_token", length: 50},
          {name: "secret_key_base", length: 50},
          {group: "saml", name: "client_id", length: 100, optional: true},
          {group: "saml", name: "client_secret", length: 100, optional: true},
          {group: "recaptcha", name: "site_key", length: 100, optional: true},
          {group: "recaptcha", name: "secret_key", length: 100, optional: true},
        ].each do |secret|
          add_secret(secret, !secret[:optional])
          remove_secret_from_hash(secret)
        end

        credentials.save

        super
      end

      def remove_secret_from_hash(secret_spec)
        group = secret_spec[:group]
        name = secret_spec[:name]

        if group && self[group]
          self[group].delete(name.to_sym)
        elsif has_key?(name.to_sym)
          delete(name.to_sym)
        end
      end

      def add_secret(secret_spec, create_if_missing = true)
        group = secret_spec[:group]
        name = secret_spec[:name]

        config_value = group ? self[group][name] : self[name]

        # use default group for ungrouped credentials
        group ||= "manage"

        if config_value
          warn_if_cred_mismatch(secret_spec)
          credentials.add(group, name, value: config_value, frozen: true, force: true)
        elsif create_if_missing
          credentials.add(group, name, length: secret_spec[:length], frozen: secret_spec[:frozen])
        end
      end

      def warn_if_cred_mismatch(opts)
        group = opts[:group]
        name = opts[:name]
        return if !credentials.exist?(group, name)

        pass_in_config = opts[:config_value] || self[group][name]
        pass_in_secrets = opts[:secrets_value] || credentials.get(group, name)
        command_name = opts[:set_command] || "set-secret #{group} #{name}"
        config_key_desc = opts[:config_key_desc] || "#{group}['#{name}']"

        if pass_in_secrets != pass_in_config
          warning = <<WARN
#{config_key_desc} in secrets store does not match the value
configured in chef-manage.rb or secrets.rb -- overriding secrets store
password with configuration file password.

If this is unexpected, consider removing the secret from those files
and setting the correct value with:

    chef-server-ctl #{command_name}
WARN
        elsif pass_in_config
          warning = <<WARN
#{config_key_desc} is configured in plain in those files -- remove
it to have the value provided from the secret store instead.
WARN
        end

        # We do better in chef-server, but this is good enough for now
        Chef::Log.warn warning if warning
      end

      def credentials
        config = {
          provider: 'chef-secrets-file',
          path: '/etc/opscode/private-chef-secrets.json'
        }
        @credentials ||= Veil::CredentialCollection::from_config(config)
      end
    end
  end
end
