#
# Cookbook Name:: omnibus-chef-manage
# Recipe:: config
#
# Copyright:: Copyright (c) 2013-2015 Chef Software, Inc.
# License:: Apache License, Version 2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

require 'yaml'

class Chef::Recipe
  include Manage::Helpers
end

class Chef::Resource::File
  include Manage::Helpers
end

config_path = '/etc/chef-manage'
config_file = "#{config_path}/manage.rb"

# Directory for config file
directory config_path do
  owner node['chef-manage']['user']['username']
  mode '0700'
end.run_action(:create)

# Config File Handling
#
# Older versions used /etc/opscode-manage/manage.rb. We now use
# /etc/chef-manage/manage.rb. This process will handle the old and new files and
# do the right thing.
#
# Loop throught these to do the same thing for secrets
chef_manage_rb = Manage::ConfigFile.new "/etc/chef-manage/manage.rb"
opscode_manage_rb = Manage::ConfigFile.new "/etc/opscode-manage/manage.rb"
chef_manage_secrets = Manage::ConfigFile.new "/etc/chef-manage/secrets.rb"
opscode_manage_secrets = Manage::ConfigFile.new "/etc/opscode-manage/secrets.rb"

[[chef_manage_rb, opscode_manage_rb],
 [chef_manage_secrets, opscode_manage_secrets]].each do |files|
  new_file, old_file = files

  # Things are a bit tricky here, because there are multiple scenarios.  But the
  # upshot is, if we have a opscode-manage/manage.rb and don't have (or have an empty)
  # chef-manage/manage.rb, then copy it over and link back.  Otherwise warn.
  if old_file.regular_file? && new_file.non_empty?
    Chef::Log.warn "#{old_file.path} is deprecated and should be removed. Using #{new_file.path}"
  elsif old_file.regular_file?
    Chef::Log.warn "Moving to #{new_file.path} for configuration - #{old_file.path} is deprecated."
    FileUtils.mv old_file.path, new_file.path
    FileUtils.ln_s new_file.path, old_file.path
  end
end

# Read in config file
begin
  Manage::Config.from_file(config_file)
# Warn if it doesn't exist
rescue Errno::ENOENT => e
  Chef::Log.debug("Missing #{config_file}. Skipping.")
rescue NoMethodError => e
  Chef::Log.fatal("Invalid attributes in #{config_file}")
  raise
end

# this is originally run in default.rb, but we may have loaded a new user from the config now
private_chef_addon 'chef-manage' do
  action :create
end

secrets_file = "#{config_path}/secrets.rb"

if ::File.exists?(secrets_file)
  Chef::Log.info("Reading secrets from file #{secrets_file}")
  begin
    Manage::Config.from_file(secrets_file)
  rescue NoMethodError => e
    Chef::Log.fatal("Invalid attributes in #{secrets_file}")
    raise
  end
end

# Manage::Config.migrate_and_validate_secrets
node.consume_attributes({ 'chef-manage' => Manage::Config.save(true) })

if File.exists?("/etc/opscode/chef-server-running.json")
  private_chef = JSON.parse(IO.read("/etc/opscode/chef-server-running.json"))
  node.consume_attributes({"private_chef" => private_chef['private_chef']})
end

# Set the fqdn to the same as the server's api_fqdn if not previously set in
# the configuration.
if node['chef-manage']['fqdn'].nil?
  node.default['chef-manage']['fqdn'] = node['private_chef']['lb']['api_fqdn']
end

# URLs for the Chef server: The "internal" URL is what we use for communicating
# from Manage to the Chef server. This is pretty much always localhost, with
# the port and protocol the Chef server is using. The "external" URL is what
# can be used to reach the Chef server from the outside, used, for example, in
# the config.rb that ships with the Starter Kit.
node.default['chef-manage']['chef_server_external_url'] =
  chef_server_external_url(node)
node.default['chef-manage']['chef_server_internal_url'] =
  chef_server_internal_url(node)

# origin is an alias for fqdn
node.default['chef-manage']['origin'] = node['chef-manage']['fqdn']

# Set the public port to the one used for the API if it's not currently set
if node['chef-manage']['public_port'].nil?
  # We haven't set the public port, and are running on a different host than
  # the API, default to 443.
  if node['chef-manage']['fqdn'] != node['private_chef']['lb']['api_fqdn']
    node.default['chef-manage']['public_port'] = 443
  # Otherwise, use the same port as the API
  else
    node.default['chef-manage']['public_port'] =
      node['private_chef']['nginx']['ssl_port']
  end
end

# The LDAP settings are derived from the Chef server settings
ldap = node['private_chef']['ldap']

# The "API":
#
# If the ldap config isn't there => not enabled
# If the ldap config is there and enabled is set, use that
# If the ldap config is there and enabled is not set, assume enabled
#
ldap_enabled = if ldap.nil? || ldap.empty? # Disabled, Chef Server pre-12.15
                 false
               elsif !ldap['enabled'].nil? # Enabled/Disabled, Chef Server 12.15+
                 ldap['enabled']
               else                        # Enabled, Chef Server pre-12.15.0
                 true
               end

node.default['chef-manage']['ldap'] = {
  'enabled'   => ldap_enabled,
  'adjective' => ldap['system_adjective']
}

# Aliases
#
# Manage internally, in various versions, uses some different names for values,
# which we ensure are defined here.

node.default['chef-manage']['external']['port'] =
  node['chef-manage']['public_port']
node.default['chef-manage']['platform']['public_url'] =
  node['chef-manage']['chef_server_external_url']
node.default['chef-manage']['platform']['url'] =
  node['chef-manage']['chef_server_internal_url']

# Strict Search Result ACLs
#
# If the server restricts search results to only those entities that
# the querier is allowed to view according to ACLs, then tell manage
# that it needn't do redundant acl checks on them.
node.default['opscode-manage']['webapp']['strict_search_result_acls'] =
  node['private_chef']['opscode-erchef']['strict_search_result_acls']

# Newrelic config
file ::File.join(node['chef-manage']['config_path'], 'newrelic.yml') do
  owner node['chef-manage']['user']['username']
  mode '0600'
  content(
    {
      'development' => {
        'monitor_mode' => false,
      },
      'production' => {
        'monitor_mode' => false,
      },
    }.to_yaml)
end

# Config file
file ::File.join(node['chef-manage']['config_path'], 'settings.yml') do
  owner node['chef-manage']['user']['username']
  mode '0600'
  # When serializing node attributes to a hash, it doesn't deep serialize all of
  # the component Mashes. The YAML comes out with each of those lines having
  # "!ruby/hash:Chef::Node::ImmutableMash" on them. When the YAML parser used by
  # Rails tries to parse these, it fails because it's expecting regular hashes.
  #
  # With this one weird trick, we convert the hash to a JSON string and then
  # parse it. This gives us a normal hash.
  content(JSON.parse(node['chef-manage'].to_json).to_yaml)
  notifies :restart, 'component_runit_service[web]' if should_notify?('web')
  notifies :restart, 'component_runit_service[worker]' if should_notify?('worker')
end

link ::File.join(node['chef-manage']['install_path'], 'embedded/service/chef-manage/config/settings/production.yml') do
  to ::File.join(node['chef-manage']['config_path'], 'settings.yml')
end

link ::File.join(node['chef-manage']['install_path'], 'embedded/service/chef-manage/config/newrelic.yml') do
  to ::File.join(node['chef-manage']['config_path'], 'newrelic.yml')
end

link ::File.join(node['chef-manage']['install_path'], 'embedded/service/chef-manage/tmp') do
  to node['chef-manage']['tmp_path']
end

directory ::File.join(node['chef-manage']['var_path'], 'run') do
  mode '0700'
  owner node['chef-manage']['user']['username']
end
