define (require) ->
  $ = require('jquery')
  AboutModal = require('views/controls/modals/about')
  AdminController = require('controllers/admin')
  Backbone = require('b')
  BrowserVersionModal = require('views/application/modals/browser_version')
  JobQueue = require("services/job_queue")
  LicenseInfo = require('models/license_info')
  Mainnav = require('views/application/mainnav')
  NoOrganizationsModal = require('views/organizations/modals/no_organizations')
  NodesParentController = require('controllers/nodes_parent')
  Organizations = require('collections/organizations')
  PoliciesController = require('controllers/policies')
  ReportFeedback = require('views/controls/modals/report_feedback')
  ReportingController = require('controllers/reporting')
  router = require('router')
  SplitPane = require('views/application/split_pane')
  UpgradeBannerView = require('views/application/upgrade_banner')
  User = require('models/user')
  Usernav = require('views/application/usernav')
  _ = require('underscore')
  applicationTpl = require('j!application/application')
  config = require('config')
  state = require('state')

  class Application extends Backbone.Layout
    template: applicationTpl
    events:
      'click #logo': 'clickLogo'
      'click .leave-feedback': 'reportFeedback'
      'click #about-manage': 'aboutModal'

    initialize: ->
      # Setup CSRF header injection
      @csrf_param = $('meta[name="csrf-param"]').attr('content')
      @csrf_token = $('meta[name="csrf-token"]').attr('content')
      $.ajaxSetup
        beforeSend: (xhr, settings) =>
          if not settings.crossDomain and settings.type != 'GET' and @csrf_token
            xhr.setRequestHeader('X-CSRF-Token', @csrf_token)

      # When an error happens, check to see if it's a 401 Unauthorized. If it
      # is, that means our session has expired and we can send the user back to
      # the sign in page.
      $(document).ajaxError (event, jqXHR) ->
        if jqXHR.status == 401
          $.ajax(
            url: '/logout'
            type: 'POST'
          ).always(-> window.location.reload())

      # Router used to dispatch URLs in the client
      router.route(/(.*?)\/+(\?.*)?$/,
        'trailFix',
        (url, params) -> @navigate(@toFragment(url, params), { trigger: true }))
      router.route('', 'root', => _.defer(_.bind(@rootRoute, @)))
      # Core selection collection, holds the objects selected in the active list view
      @selection = new Backbone.Collection

      @splitPane = new SplitPane({app: @, selection: @selection})

      # master queue for background requests
      @masterQueue = new JobQueue({storeErrorQueue: true})
      @setUpNavigationPrevention(@masterQueue)

      # Super hacky way to expose app to organization object,
      # but not sure how else to do it and I tried for several hours
      for org in (config.organizations || [])
        org.app = @

      # Master collection for available organizations
      @organizations = new Organizations(config.organizations)
      @organizations.on('current:changed', (org) => @currentOrg = org)
      @currentOrg = @organizations.last()
      window.Chef?.preload?(@organizations)

      # Master object for current user
      state.set({ user: new User(config.user, collection: @currentOrg?.users) })

      $(window).on('resize', $.proxy(@onResize, @))

      unless @organizations.isEmpty()
        @mainTabs = [
          new NodesParentController({app: @}),
          new ReportingController({app: @}),
          new PoliciesController({app: @}),
          new AdminController({app: @}),
        ]
        @usernav = new Usernav({app: @})
        @mainnav = new Mainnav(el: false, app: @)
        @setViews({'#mainnav': @mainnav, '#usernav': @usernav})

    serialize: -> {
        support_site_url: config.support_site_url,
        privacy_policy_url: config.privacy_policy_url,
        cookie_policy_url: config.cookie_policy_url
      }

    afterRender: ->
      @onResize()
      Backbone.history.start({pushState: true})
      # If needed, display the version warning
      new BrowserVersionModal({app: @})

      # Show the banner view if needed
      unless config.disable_upgrade_banner
        @licenseInfo = new LicenseInfo
        @upgradeBannerView = new UpgradeBannerView({
          el: @$('#banner')
          model: @licenseInfo
        })
        @listenTo(@upgradeBannerView, 'afterRender', @onResize)
        @licenseInfo.fetch()

      if @organizations.isEmpty()
        new NoOrganizationsModal({app: @})

    cleanup: -> $(window).off('resize', $.proxy(@onResize, @))

    # basically here to get around circular dependencies
    goToNoOrganizationsModal: ->
      new NoOrganizationsModal({app: @})

    onResize: ->
      @splitPane.resizePanes()
      @$('#content').css('top', @$el.height())

    applyController: (controller, splitPane, paneView) ->
      state.set({ controller })
      # Set the class on the body element for any systemic CSS changes
      $('body').attr('class', controller.id)
      # Update the nav bars
      @mainnav.applyController(controller)
      @usernav.applyController(controller)
      # Request the first page of items if needed
      if controller.collection?.url and controller.collection.length == 0
        controller.collection.fetch()
      # Populate the sidebar
      newSidebar = controller.sidebarView()
      unless newSidebar is @sidebarView
        @sidebarView = newSidebar
        @setView('#sidebar', @sidebarView).render()
      # Set the correct sidebar section
      for section, i in @sidebarView.sections
        if section.controller == controller
          # Open the section, but don't trigger events. The silent makes it so
          # we do not navigate. We don't need to, we're already here.
          @sidebarView.openThisAndCloseAllOthers(@$(section.el), { silent: true })

      # Figure out what pane view to use and install it
      if splitPane ? controller.split_pane
        unless @paneView is @splitPane
          @paneView = @splitPane
          @setView('#panes', @splitPane).render()
        @splitPane.applyController(controller)
      else
        @paneView = paneView ? _.result(controller, 'paneView')
        @setView('#panes', @paneView).render()
        @splitPane.listViewClass = undefined
        @splitPane.objectViewClass = undefined
        @splitPane.activeObjectView = undefined

    rootRoute: ->
      unless @organizations.isEmpty()
        for tab in @mainTabs
          if tab.checkAccess()
            defaultTab = tab
            break
        if defaultTab
          router.navigate(defaultTab.url(), { trigger: true })
        else
          # This should be impossible
          router.navigate(@organizations.url(), { trigger: true })

    clickLogo: (evt) ->
      evt.preventDefault()
      @rootRoute()

    aboutModal: (evt) ->
      evt.preventDefault()
      new AboutModal({app: @, collection: null})

    reportFeedback: (evt) ->
      evt.preventDefault()
      new ReportFeedback({ app: @, controller: state.get('controller') })

    # Check if we're on hosted by using darklaunch
    isHosted: -> not config.private_chef

    setUpNavigationPrevention: (masterQueue) ->
      (window.Chef || (window.Chef = {})).masterQueue = masterQueue

      window.onbeforeunload = ->
        if window.Chef.masterQueue.length() != 0
          'You have background requests still waiting to be sent to the server for processing! If you leave the console, these requests will not be finished.'
