define (require) ->
  Backbone = require("b")

  # See full documentation in doc/ark.md#services
  # input:
  #   options:
  #     requestArray [
  #       type (required): HTTP verb for ajax call
  #       url (required): url for ajax call
  #       data (optional): data sent to server by ajax call
  #       model (optional): the backbone model for this request
  #       successEvent (optional): method with input (data, status, xhr, model) that handles a successful request
  #       errorEvent (optional): method with input (xhr, status, error, model) that handles a failed request
  #     ]
  # signals:
  #   success (responseData, status, xhr, model): a request has succeeded
  #   error (xhr, status, error, request.model): a request has failed
  #   all_requests_sent (no args): triggers when @processQueue() has sent all events to server
  #   queue_extended (no args): triggers when someone extends the queue with another queue

  # only extends Backbone.Model to use Backbone's eventing system
  class JobQueue extends Backbone.Model
    # ============== #
    # public methods #
    # ============== #

    initialize: (options) ->
      # =================== #
      # configurable fields #
      # =================== #
      @requestArray = if options?.requestArray? then options.requestArray else []

      # flag for whether to store failed requests in an error queue, defaults to false
      @storeErrorQueue = if options?.storeErrorQueue? then options.storeErrorQueue else false

      # ============== #
      # private fields #
      # ============== #

      # whether or not @processQueue() is running or not
      @_processingJobs = false

      # array of errors, exposed externally by @generateErrorQueue()
      @_errorArray = []

      # will be true when we want to terminate a running @processQueue()
      @_sendToBackground = false

    # recursive method that will process the queue until
    # it is empty or the signal has been sent to send
    # remaining requests to the new queue
    processQueue: ->
      if @requestArray.length != 0
        @_processingJobs = true
        if not @_sendToBackground
          request = @requestArray.splice(0,1)[0]
          @_sendNextRequest(request)
        else
          @_terminateAndSendToOtherQueue()
      else
        @_processingJobs = false
        @trigger("all_requests_sent")

    # send the remaining requests to a different JobQueue
    sendRemainingToDifferentQueue: (jobQueue) ->
      @otherJobQueue = jobQueue
      # catch on next iteration during @processQueue()
      if @_processingJobs
        @_sendToBackground = true
      # @processQueue() isn't running, just call termainate directly instead
      # of letting @processQueue() pick it up
      else
        @_terminateAndSendToOtherQueue()

    removeIndex: (index) ->
      if @_processingJobs == true
        throw new Error("You cannot remove a specific index while the queue is being processed.")
      @requestArray.splice(index,1)[0]

    queueRequest: (request) ->
      @requestArray.push(request)
      @trigger("queue_extended")

    length: ->
      @requestArray.length

    generateErrorQueue: ->
      if not @storeErrorQueue
        throw new Error("You did not elect to store errors in this queue. To store them, set @storeErrorQueue to true on initialization.")
      else
        new JobQueue({requestArray: @_errorArray})

    numberOfErrors: ->
      if not @storeErrorQueue
        throw new Error("You did not elect to store errors in this queue. To store them, set @storeErrorQueue to true on initialization.")
      else
        @_errorArray.length

    # =============== #
    # private methods #
    # =============== #
    # no need to call externally

    # send remaining requests to new queue and start it up
    _terminateAndSendToOtherQueue: ->
      # send remaining jobs to other queue
      @otherJobQueue._extendQueue(@)
      @requestArray = []

      # and process new jobs if the queue isn't running
      if not @otherJobQueue._processingJobs
        @otherJobQueue.processQueue()
      @trigger("all_requests_sent")

    # dump a different JobQueue at the end of this queue
    _extendQueue: (jobQueue) ->
      @requestArray.push.apply(@requestArray, jobQueue.requestArray)
      @trigger("queue_extended")

    # submits a request to the server, with error and suceess event handling,
    # as well as adding errors to @_errorArray
    _sendNextRequest: (request) ->
      $.ajax
        type: request.type
        url: request.url
        data: request.data
        success: (responseData, status, xhr) =>
          request.successEvent(responseData, status, xhr, request.model) if request.successEvent?
          @trigger("success", responseData, status, xhr, request.model)
          @processQueue()
        error: (xhr, status, error) =>
          request.errorEvent(xhr, status, error, request.model) if request.errorEvent?
          if @storeErrorQueue
            request.originalError = {
              xhr: xhr,
              status: status,
              error: error
            }
            @_errorArray.push(request)
          @trigger("error", xhr, status, error, request.model)
          @processQueue()
