# Forked from https://github.com/thedersen/backbone.validation
# Original content is copyright Thomas Pedersen under the MIT license
define ['backbone', 'underscore', 'jquery'], (Backbone, _, $) ->
  # Helper functions
  # ----------------

  # Formatting functions used for formatting error messages
  formatFunctions =
    # Uses the configured label formatter to format the attribute name
    # to make it more readable for the user
    formatLabel: (attrName, model) ->
      Validation.labelFormatters[Validation.options.labelFormatter](attrName, model)


    # Replaces nummeric placeholders like {0} in a string with arguments
    # passed to the function
    format: (text, args...) ->
      text.replace /\{(\d+)\}/g, (match, number) ->
        args[number] ? match


  # Flattens an object
  # eg:
  #
  #     {
  #       address: {
  #         street: 'Street',
  #         zip: 1234
  #       }
  #     };
  #
  # becomes:
  #
  #     {
  #       'address.street': 'Street',
  #       'address.zip': 1234
  #     };
  flatten = (obj, into, prefix) ->
    into = into or {}
    prefix = prefix or ""
    for key, val of obj
      if obj.hasOwnProperty(key)
        if val and typeof val is "object" and not (val instanceof Date or val instanceof RegExp)
          flatten(val, into, prefix + key + ".")
        else
          into[prefix + key] = val

    into


  # Use native trim when defined
  trim = if String::trim
    (text) ->
      if text is null then "" else String::trim.call(text)
  else
    (text) ->
      if text is null then "" else text.toString().replace(/^\s+|\s+$/g, '')


  # Determines whether or not a value is a number
  isNumber = (value) ->
    _.isNumber(value) or (_.isString(value) and value.match(Validation.patterns.number))


  # Determines whether or not not a value is empty
  hasValue = (value) ->
    value? and trim(value) != ''


  # Returns an object with undefined properties for all
  # attributes on the model that has defined one or more
  # validation rules.
  getValidatedAttrs = (model) ->
    attrs = {}
    for key of model.validation or {}
      attrs[key] = undefined
    attrs


  # Looks on the model for validations for a specified
  # attribute. Returns an array of any validators defined,
  # or an empty array if none is defined.
  getValidators = (model, attr) ->
    attrValidationSet = model.validation?[attr] ? {}

    # If the validator is a function or a string, wrap it in a function validator
    if _.isFunction(attrValidationSet) or _.isString(attrValidationSet)
      attrValidationSet = {fn: attrValidationSet}

    # Stick the validator object into an array
    unless _.isArray(attrValidationSet)
      attrValidationSet = [attrValidationSet]

    # Reduces the array of validators into a new array with objects
    # with a validation method to call, the value to validate against
    # and the specified error message, if any
    validators = []
    for attrValidation in attrValidationSet
      for validator of attrValidation
        continue if validator == 'msg' # Special key
        validators.push({
          fn: Validation.validators[validator],
          val: attrValidation[validator],
          msg: attrValidation.msg,
        })
    validators


  # Validates an attribute against all validators defined
  # for that attribute. If one or more errors are found,
  # the first error result is returned.
  # If the attribute is valid, an empty string promise is returned.
  validateAttr = (model, attr, value, computed) ->
    dfds = []
    for validator in getValidators(model, attr)
      # Pass the format functions plus the default
      # validators as the context to the validator
      ctx = _.extend({}, formatFunctions, Validation.validators)
      result = validator.fn.call(ctx, value, attr, validator.val, model, computed)
      if result == false
        return $.when(false) # Early bail out
      dfds.push if not result or _.isFunction(result.promise)
        result
      else
        new $.Deferred (dfd) ->
          dfd.reject(validator.msg or result)
    $.when(dfds...)


  # Loops through the model's attributes and validates them all.
  # Returns and object containing names of invalid attributes
  # as well as error messages.
  validateModel = (model, attrs) ->
    attrDfds = {}
    for attr, val of flatten(attrs)
      attrDfds[attr] = validateAttr(model, attr, val, attrs)
    dfd = $.when(_.values(attrDfds)...)
    dfd.attributes = attrDfds
    dfd


  # Contains the methods that are mixed in on the model when binding
  mixin = (view, options) ->

    # Check whether or not a value passes validation
    # without updating the model
    preValidate: (attr, value) ->
      validateAttr(this, attr, value, _.extend({}, @attributes))


    # Check to see if an attribute, an array of attributes or the
    # entire model is valid. Passing true will force a validation
    # of the model.
    isValid: (option) ->
      flattened = flatten(@attributes)
      if _.isString(option)
        return validateAttr(this, option, flattened[option], _.extend({}, @attributes))
      if _.isArray(option)
        dfds = for attr in option
          validateAttr(this, attr, flattened[attr], _.extend({}, @attributes))
        return $.when(dfds...)
      @validate() if option == true
      if @validation then @_isValid else true


    # This is called by Backbone when it needs to perform validation.
    # You can call it manually without any parameters to validate the
    # entire model.
    validate: (attrs, setOptions) ->
      throw 'Do not use validate:true with .set()/.save()' if setOptions?.validate
      model = this
      validateAll = not attrs
      opt = _.extend({}, options, setOptions)
      validatedAttrs = getValidatedAttrs(model)
      allAttrs = _.extend({}, validatedAttrs, model.attributes, attrs)
      changedAttrs = flatten(attrs or allAttrs)
      result = validateModel(model, allAttrs)
      model._isValid = result

      # Install callbacks on the deferreds
      for attr, val of validatedAttrs
        dfd = result.attributes[attr]
        if dfd.state() == 'pending'
          opt.validating.call(model, view, attr, opt.selector)
        do (attr, val) ->
          dfd.done (response) ->
            opt.valid.call(model, view, attr, opt.selector)
          if validateAll or changedAttrs.hasOwnProperty(attr)
            dfd.fail (response) ->
              opt.invalid.call(model, view, attr, response, opt.selector)

      # Trigger validated events.
      # Need to defer this so the model is actually updated before
      # the event is triggered, in case all validation was sync-mode.
      _.defer ->
        result.done ->
          model.trigger('validated', model, result, true)
          model.trigger('validated:valid', model, result)
        .fail ->
          model.trigger('validated', model, result, false)
          model.trigger('validated:invalid', model, result)

      result

  # Helper to mix in validation on a model
  bindModel = (view, model, options) ->
    _.extend(model, mixin(view, options))


  # Removes the methods added to a model
  unbindModel = (model) ->
    delete model.validate
    delete model.preValidate
    delete model.isValid


  # Mix in validation on a model whenever a model is
  # added to a collection
  collectionAdd = (model) ->
    bindModel(@view, model, @options)


  # Remove validation from a model whenever a model is
  # removed from a collection
  collectionRemove = (model) ->
    unbindModel(model)


  Validation = Backbone.Validation =
    # Current version of the library
    version: '0.7.1'

    # Default options
    options:
      forceUpdate: false
      selector: 'name'
      labelFormatter: 'sentenceCase'
      invalid: Function.prototype
      valid: Function.prototype
      validating: Function.prototype

      # Called to configure the default options
      configure: (options) ->
        _.extend(Validation.options, options)

    # Hooks up validation on a view with a model
    # or collection
    bind: (view, options) ->
      model = view.model
      collection = view.collection
      options = _.extend({}, Validation.options, Validation.callbacks, options)
      throw """
            Before you execute the binding your view must have a model or a collection.
            See http://thedersen.com/projects/backbone-validation/#using-form-model-validation for more information.
            """ unless model ? collection
      if model
        bindModel(view, model, options)
      else if collection
        for model in collection.models
          bindModel(view, model, options)
        collection.bind('add', collectionAdd, {view: view, options: options})
        collection.bind('remove', collectionRemove)

    # Removes validation from a view with a model
    # or collection
    unbind: (view) ->
      model = view.model
      collection = view.collection
      if model
        unbindModel(model)
      if collection
        for model in collection.models
          unbindModel(model)
        collection.unbind('add', collectionAdd)
        collection.unbind('remove', collectionRemove)

    # Used to extend the Backbone.Model.prototype
    # with validation
    mixin: ->
      mixin(null, Validation.options)

    # Callbacks
    # ---------
    callbacks:

      # Gets called at the start of validation
      validating: (view, attr, selector) ->

      # Gets called when a previously invalid field in the
      # view becomes valid. Removes any error message.
      # Should be overridden with custom functionality.
      valid: (view, attr, selector) ->
        view.$("[" + selector + "~=\"" + attr + "\"]").removeClass("invalid").removeAttr "data-error"

      # Gets called when a field in the view becomes invalid.
      # Adds a error message.
      # Should be overridden with custom functionality.
      invalid: (view, attr, error, selector) ->
        view.$("[" + selector + "~=\"" + attr + "\"]").addClass("invalid").attr "data-error", error


    # Patterns
    # --------
    patterns:

      # Matches any digit(s) (i.e. 0-9)
      digits: /^\d+$/

      # Matched any number (e.g. 100.000)
      number: /^-?(?:\d+|\d{1,3}(?:,\d{3})+)(?:\.\d+)?$/

      # Matches a valid email address (e.g. mail@example.com)
      email: /^((([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+(\.([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+)*)|((\x22)((((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(([\x01-\x08\x0b\x0c\x0e-\x1f\x7f]|\x21|[\x23-\x5b]|[\x5d-\x7e]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(\\([\x01-\x09\x0b\x0c\x0d-\x7f]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))))*(((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(\x22)))@((([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.)+(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))$/i

      # Mathes any valid url (e.g. http://www.xample.com)
      url: /^(https?|ftp):\/\/(((([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(%[\da-f]{2})|[!\$&'\(\)\*\+,;=]|:)*@)?(((\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5])\.(\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5])\.(\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5])\.(\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5]))|((([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.)+(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.?)(:\d*)?)(\/((([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(%[\da-f]{2})|[!\$&'\(\)\*\+,;=]|:|@)+(\/(([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(%[\da-f]{2})|[!\$&'\(\)\*\+,;=]|:|@)*)*)?)?(\?((([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(%[\da-f]{2})|[!\$&'\(\)\*\+,;=]|:|@)|[\uE000-\uF8FF]|\/|\?)*)?(\#((([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(%[\da-f]{2})|[!\$&'\(\)\*\+,;=]|:|@)|\/|\?)*)?$/i


    # Error messages
    # --------------

    # Error message for the build in validators.
    # {x} gets swapped out with arguments form the validator.
    messages:
      required: "{0} is required"
      acceptance: "{0} must be accepted"
      min: "{0} must be greater than or equal to {1}"
      max: "{0} must be less than or equal to {1}"
      range: "{0} must be between {1} and {2}"
      length: "{0} must be {1} characters"
      minLength: "{0} must be at least {1} characters"
      maxLength: "{0} must be at most {1} characters"
      rangeLength: "{0} must be between {1} and {2} characters"
      oneOf: "{0} must be one of: {1}"
      equalTo: "{0} must be the same as {1}"
      pattern: "{0} must be a valid {1}"


    # Label formatters
    # ----------------

    # Label formatters are used to convert the attribute name
    # to a more human friendly label when using the built in
    # error messages.
    # Configure which one to use with a call to
    #
    #     Backbone.Validation.configure({
    #       labelFormatter: 'label'
    #     });
    labelFormatters:

      # Returns the attribute name with applying any formatting
      none: (attrName) ->
        attrName


      # Converts attributeName or attribute_name to Attribute name
      sentenceCase: (attrName) ->
        attrName.replace(/(?:^\w|[A-Z]|\b\w)/g, (match, index) ->
          (if index is 0 then match.toUpperCase() else " " + match.toLowerCase())
        ).replace "_", " "


      # Looks for a label configured on the model and returns it
      #
      #      var Model = Backbone.Model.extend({
      #        validation: {
      #          someAttribute: {
      #            required: true
      #          }
      #        },
      #
      #        labels: {
      #          someAttribute: 'Custom label'
      #        }
      #      });
      label: (attrName, model) ->
        (model.labels and model.labels[attrName]) or Validation.labelFormatters.sentenceCase(attrName, model)


    # Built in validators
    # -------------------
    validators:

      # Function validator
      # Lets you implement a custom function used for validation
      fn: (value, attr, fn, model, computed) ->
        fn = model[fn]  if _.isString(fn)
        fn.call model, value, attr, computed


      # Required validator
      # Validates if the attribute is required or not
      required: (value, attr, required, model, computed) ->
        isRequired = (if _.isFunction(required) then required.call(model, value, attr, computed) else required)
        return false  if not isRequired and not hasValue(value) # overrides all other validators
        @format Validation.messages.required, @formatLabel(attr, model)  if isRequired and not hasValue(value)


      # Acceptance validator
      # Validates that something has to be accepted, e.g. terms of use
      # `true` or 'true' are valid
      acceptance: (value, attr, accept, model) ->
        @format Validation.messages.acceptance, @formatLabel(attr, model)  if value isnt "true" and (not _.isBoolean(value) or value is false)


      # Min validator
      # Validates that the value has to be a number and equal to or greater than
      # the min value specified
      min: (value, attr, minValue, model) ->
        @format Validation.messages.min, @formatLabel(attr, model), minValue  if not isNumber(value) or value < minValue


      # Max validator
      # Validates that the value has to be a number and equal to or less than
      # the max value specified
      max: (value, attr, maxValue, model) ->
        @format Validation.messages.max, @formatLabel(attr, model), maxValue  if not isNumber(value) or value > maxValue


      # Range validator
      # Validates that the value has to be a number and equal to or between
      # the two numbers specified
      range: (value, attr, range, model) ->
        @format Validation.messages.range, @formatLabel(attr, model), range[0], range[1]  if not isNumber(value) or value < range[0] or value > range[1]


      # Length validator
      # Validates that the value has to be a string with length equal to
      # the length value specified
      length: (value, attr, length, model) ->
        @format Validation.messages.length, @formatLabel(attr, model), length  if not hasValue(value) or trim(value).length isnt length


      # Min length validator
      # Validates that the value has to be a string with length equal to or greater than
      # the min length value specified
      minLength: (value, attr, minLength, model) ->
        @format Validation.messages.minLength, @formatLabel(attr, model), minLength  if not hasValue(value) or trim(value).length < minLength


      # Max length validator
      # Validates that the value has to be a string with length equal to or less than
      # the max length value specified
      maxLength: (value, attr, maxLength, model) ->
        @format Validation.messages.maxLength, @formatLabel(attr, model), maxLength  if not hasValue(value) or trim(value).length > maxLength


      # Range length validator
      # Validates that the value has to be a string and equal to or between
      # the two numbers specified
      rangeLength: (value, attr, range, model) ->
        @format Validation.messages.rangeLength, @formatLabel(attr, model), range[0], range[1]  if not hasValue(value) or trim(value).length < range[0] or trim(value).length > range[1]


      # One of validator
      # Validates that the value has to be equal to one of the elements in
      # the specified array. Case sensitive matching
      oneOf: (value, attr, values, model) ->
        @format Validation.messages.oneOf, @formatLabel(attr, model), values.join(", ")  unless _.include(values, value)


      # Equal to validator
      # Validates that the value has to be equal to the value of the attribute
      # with the name specified
      equalTo: (value, attr, equalTo, model, computed) ->
        @format Validation.messages.equalTo, @formatLabel(attr, model), @formatLabel(equalTo, model)  if value isnt computed[equalTo]


      # Pattern validator
      # Validates that the value has to match the pattern specified.
      # Can be a regular expression or the name of one of the built in patterns
      pattern: (value, attr, pattern, model) ->
        @format Validation.messages.pattern, @formatLabel(attr, model), pattern  if not hasValue(value) or not value.toString().match(Validation.patterns[pattern] or pattern)
