define (require) ->
  Backbone = require('b')
  BasePaginator = require('collections/base')
  Dropdown = require('views/controls/dropdown')
  Empty = require('views/controls/empty')
  ErrorAlertView = require('views/controls/alerts/error')
  Grid = require('views/controls/grid')
  ListHeader = require('views/controls/list_header')
  Organizations = require('collections/organizations')
  ScrollGrid = require('views/controls/scroll_grid')
  Spinner = require('views/controls/spinner')
  T = require('utils/i18n.gettext')
  _ = require('underscore')
  actionTemplate = require('j!controls/list_action')
  router = require('router')
  state = require('state')
  template = require('j!controls/list')

  class ListView extends Backbone.Layout
    paginated: true
    template: template
    columns: []
    actionColumn: {
      id: "actions",
      name: T("Actions")
      width: 70,
      resizable: false,
      cssClass: "actions",
      formatter: (row, cell, value, columnDef, dataContext) ->
        actionTemplate({ cid: dataContext.cid })
    }
    gridOptions: {}
    gridClass: Grid
    scrollGridClass: ScrollGrid
    headerClass: ListHeader
    customHeaderClass: '.list-header'

    events:
      'click .dropdown-toggle': 'toggleActionsMenu'

    initialize: (@options = {}) ->
      @app = @options.app
      @collection = @options.collection ? new Backbone.Collection
      @controller = @options.controller
      @gridPadding = 5
      @selection = @options.selection ? new Backbone.Collection

      gridOptions = _.result(@, 'gridOptions')
      # Force this on
      gridOptions.explicitInitialization = true
      # Check for .paginator_core to know if this is a paginated collection
      gridClass = if @collection.paginator_core
        @scrollGridClass
      else
        @gridClass

      columns = _.result(@, 'columns')
      columns = columns.concat(_.result(@, 'actionColumn')) if @actionColumn

      @grid = new gridClass({
        collection: @collection
        selection: @selection
        columns: columns
        gridOptions: gridOptions
      })

      @listHeader = new @headerClass({
        app: @app
        collection: @collection
        listView: @
        messages: @controller?.messages
      })
      @listHeader.on("search", @searchEvent, @)
      @setViews({'.list-header': @listHeader})

      if @paginated
        @collection.on("sync error", @removeSpinner, @)
        @collection.on("error", @showError, @)
        @collection.on("request", @installSpinner, @)
      @collection.on("add", @renderIfEmpty, @)
      @collection.on("delete", @clearSelection, @)
      @collection.on("render", @renderGrid, @)

      if @collection instanceof BasePaginator
        if @collection.params?.filter?
          @collection.searchFilter = @collection.params.filter
        else
          @collection.searchFilter = "*:*"
        @listHeader.resetListAndCollectionState() if @collection.isEmpty()

    initActionsDropdown: (actions)->
      dropdownItems = _.filter(actions, (action)-> action.dropdown ? true)
      @actionsDropdown = new Dropdown(items: dropdownItems)
      @actionsDropdown.on('action', @dispatchAction, @)

    clearSelection: ->
      @selection.reset()
      @updateSelection()

    renderIfEmpty: ->
      # if the collection is now not empty but the empty message is
      # rendered, should probably re-render the grid
      if @collection.length > 0 and @$(".empty").length > 0
        @renderGrid()

    beforeRender: ->
      @selection.on('select', @updateSelection, @)

    afterRender: ->
      @resizeGrid()

      if @paginated and @collection.queryStatus?.initQuery
        @installSpinner()
      else
        @renderGrid()

    renderGrid: ->
      @listHeader.updateTitleText()

      if @collection.length == 0
        @setView('.grid', new Empty()).render()
        return

      unless @getView('.actions-dropdown')
        @initActionsDropdown(@options.actions ? [])
        @setView('.actions-dropdown', @actionsDropdown).render()

      @setView('.grid', @grid).render()

    cleanup: ->
      # the search state does not properly persist (and it would add a lot of
      # complexity to work in such a way) so just reset the collection so it
      # gets reloaded if it currently had a filter to reduce state.
      if @collection.searchFilter? != "*:*" and (not @collection instanceof Organizations) and @collection.isSolrSearchable
        @collection.reset()
        @collection.searchFilter = '*:*'

      @selection.off(null, null, @)
      @listHeader.off(null, null, @)
      @grid.off(null, null, @)
      @spinner = undefined

    installSpinner: (collection, xhr, options) ->
      if @collection.queryStatus?.initQuery
        @spinner = new Spinner
        @setView('.grid', @spinner).render()
      else if @collection.queryStatus?.queryingPageDown
        @collection.push({id: '__backbone_spinner__', isSpinner: true})

    removeSpinner: (collection, resp, options) ->
      if options.queryInitial
        @spinner?.remove()
        @spinner = undefined

        # TODO Why do we call @renderGrid in @removeSpinner?
        @renderGrid()
      else if options.queryDown
        @collection.remove(@collection.filter((m) -> m.get('isSpinner')))

    resizeGrid: ->
      # Size the header area and grid container area within list pane
      listPaneHeight = @$el.parent().height()
      headerHeight = @$(@customHeaderClass).height()
      extraPadding = @gridPadding

      @$('.list-container').height(
        listPaneHeight - (headerHeight + extraPadding))

      if @grid
        @grid.resize()

    showError: (collection, xhr, options) ->
      # if the HTTP verb was defined and it is not GET, do not handle this error,
      # as it was not a list error
      if not (xhr.responseJSON?.verb? and xhr.responseJSON.verb != "GET")

        if xhr.status == 403
          objectDescription = if state.get('controller')?
            state.get('controller').messages.label(2).toLowerCase()
          else
            "this object type"

          errorView = new ErrorAlertView({title: "Permissions Error", body: "You do not have list permissions for #{objectDescription}. Please ask an administrator for access and try again."})

        else
          errorView = new ErrorAlertView()
        @setView('.grid', errorView).render()

    toggleActionsMenu: (evt) ->
      evt.stopPropagation()
      evt.preventDefault()
      targetModel = @collection.get($(evt.currentTarget).data('cid'))
      if targetModel == @actionsDropdownCurrent
        @actionsDropdown.toggle()
        if @actionsDropdown.isOpen
          @actionsDropdown.positionAt($(evt.currentTarget))
      else
        @actionsDropdown.open()
        @actionsDropdown.positionAt($(evt.currentTarget))
        @actionsDropdownCurrent = targetModel

    updateSelection: ->
      if @selection.length <= 1
        @$el.removeClass('list-multiselect')
        url = if @selection.length == 1
                @selection.at(0).url()
              else
                @collection.urlWithQueryParams()
        router.navigate(url)
      else
        @$el.addClass('list-multiselect')

    listSelectionUpdated: ->
      @trigger('select')

    resetSelection: ->
      @grid.resetSelection()

    resetState: ->
      @grid.resetState() if @grid?.resetState?

    dispatchAction: (action) ->
      @trigger("action", action) # Proxy @actionsDropdown events so other
                                 # objects can listen to ListView instead of
                                 # breaking the LoD with
                                 # someListView.actionsDropdown.on(...)
      @controller?.dispatchAction(action)

    searchEvent: ->
      @selection.reset()
      @updateSelection()
