define (require) ->
  Backbone = require('b')
  SlickBack = require('slickback')
  Grid = require('views/controls/grid')
  AddButton = require('views/controls/add_button')
  tpl = require('j!controls/permissions')
  T = require('utils/i18n.gettext')
  _ = require('underscore')
  $ = require('jquery')
  require('bootstrap-tooltip')

  # This function is called every time a new checkbox is formed. Below we are
  # building a string that changes the checkbox behavior based on the permissions.
  checkboxFormatter = (row, cell, value, columnDef, dataContext) ->
    # The checkbox should be disabled if it represents the grant permission of
    # the admins group.
    isAdminsGrant = columnDef.field == "grant" &&
      dataContext.get("name") == "admins" &&
      dataContext.get("type") == "group"

    # Disabled input elements don't appear to reliably generate
    # mouseleave events, not even any that bubble up to the container.
    # cf. http://bugs.jquery.com/ticket/11382 and
    # https://github.com/twbs/bootstrap/issues/10049
    # A workaround is to force a div *over* the checkbox and attach the
    # tooltip there.
    tooltip = if isAdminsGrant
      """
      <div title="The 'admins' group must have Grant permissions.
                  This permission cannot be removed."
         data-toggle="tooltip" data-container="body"
         style="height: 100%; width: 80%; position: absolute;
         z-index: 100;"></div>
      """
    else
      ""

    """
    <div>
    #{tooltip}
    <input class="perm" type="checkbox"
    #{if isAdminsGrant then "disabled" else ""}
    #{if isAdminsGrant or value then "checked" else ""}
    data-cid="#{dataContext.cid}" data-perm="#{columnDef.field}"/>
    </div>
    """

  # Mapping of ACL entry type to icon, because lolconsistency
  aclIcons = {group: 'group', user: 'user', client: 'briefcase'}

  class PermissionsView extends Backbone.Layout
    template: tpl
    permissionsColumns: [
      {id: 'read', name: T('Read')}
      {id: 'update', name: T('Update')}
      {id: 'delete', name: T('Delete')}
      {id: 'grant', name: T('Grant')}
    ]
    nameColumn:
      id: 'name'
      name: T('Name')
      field: 'name'
      headerCssClass: 'name'
      cssClass: 'name'
      width: 120
      resizable: false
      formatter: (row, cell, value, columnDef, dataContext) ->
        """<i class="fa fa-#{aclIcons[dataContext.get('type')]}"></i>#{value}"""

    events:
      'keyup input.filter': 'updateFilter'
      'change input.perm': 'permissionToggle'

    initialize: ->
      @app = @options.app
      @originalCollection = @model.acl()
      # Copy the collection for use with filtering
      @collection = new SlickBack.Collection
      @collectionUpdate()
      @grid = new Grid({
        collection: @collection,
        columns: _.result(@, 'columns'),
        gridOptions: {forceFitColumns: false, autoHeight: true, rowHeight: 25},
      })
      @setView('.grid', @grid)
      # Setup the add button
      @addButton = new AddButton
      @addButton.on('add', @addEntry, @)
      @setView('span.add', @addButton)
      # Because 5ms flashes get really annoying
      @updateSavingBanner = _.throttle(@updateSavingBanner, 100)

    columns: ->
      [@nameColumn].concat(for col in @permissionsColumns
        id: col.id
        name: col.name
        field: col.id
        width: col.width or 80
        resizable: false
        formatter: checkboxFormatter
      )

    beforeRender: ->
      # Reset the filter before each render to be safe
      @filterValue = ''
      @collection.reset(@originalCollection.models)
      # Install the collection handlers
      @originalCollection.off(null, null, @)
      @originalCollection.on('reset add remove error', @collectionUpdate, @)
      @originalCollection.on('error', @render, @)

    cleanup: ->
      @originalCollection.off(null, null, @)

    afterRender: ->
      if @originalCollection.status == 403
        @$(".add-button").attr("disabled", true)

    serialize: ->
      if @originalCollection.status == 403
        errorType = "&nbsp;Permissions Error"
        errorDescription = "You do not have the permissions to view this information."
      else
        errorType = "&nbsp;Server Error"
        errorDescription = "Something went wrong. Please try again later."
      {error: @originalCollection.error, errorType: errorType, errorDescription: errorDescription}

    collectionUpdate: ->
      @applyFilter(@filterValue, true)
      @$el.toggleClass('loading-toggle', @originalCollection.length == 0 && !@originalCollection.error)
      # Activate tooltips on checkboxes in the rendered collection
      @$("[data-toggle=tooltip]").tooltip()
      @updateWarning()

    # Handles adding / removing warning tooltip and styling in the case that
    # an object has all of its permissions removed. Explains to user
    # a bit about how that permissions case works.
    updateWarning: ->
      @$(".slick-row").each( (index) ->
        # if an object has no permissions, show message
        if $(@).find(":checkbox:checked").size() == 0
          $(@).children(".name").addClass("warning").tooltip({container: 'body', title: "The Chef permissions system assumes that an object has no permissions by default, so if you just added this object, its permissions will only actually be updated if you add a permission. Also, if you just removed all permissions on this object, do not be alarmed if it is not shown here in the future."}).append("<i class='fa fa-info-sign'></i>")
        else # otherwise, remove message
          $(@).children(".name").removeClass("warning").tooltip('destroy').children(".fa-info-sign").remove()
      )

    permissionToggle: (evt) ->
      $el = @$(evt.target)
      entry = @originalCollection.get($el.data('cid'))
      ace = $el.data('perm')
      value = $el.is(':checked')
      entry.set(ace, value)
      @originalCollection.actions += 1
      @updateSavingBanner()

      $.ajax
        type: 'PUT'
        dataType: 'json'
        url: @originalCollection.url()
        data:
          ace: ace
          type: entry.get('type')
          name: entry.get('name')
          value: value
        complete: =>
          @originalCollection.actions -= 1
          @originalCollection.actions = 0 if @originalCollection.actions < 0 # Just in case of crazy
          @updateSavingBanner()
          @updateWarning()
        error: (model, resp) =>
          @$el.addClass('saving-error-toggle')
          entry.set(ace, !value)
          $el.prop('checked', !value)

    addEntry: (name) ->
      if @originalCollection.findWhere({ name })
        @addButton.error(T('Could not add duplicate entry.'))
      else
        @addButton.enableSpinner()
        @originalCollection.create(
          {name: name},
          {
            url: "#{@originalCollection.collection.organization.url()}/acls",
            wait: true,
            success: =>
              @addButton.disableSpinner().close()
            error: (model, resp) =>
              msg = if resp.status == 404
                T("Object '#{model.get('name')}' is unknown.")
              else
                T('An error occurred, please try again.')
              @addButton.error(msg).disableSpinner()
          }
        )

    updateFilter: (evt) ->
      $el = @$('input.filter')
      value = if evt.which == 27
        $el.val('')
        ''
      else
        $el.val().toLowerCase()
      @applyFilter(value)

    applyFilter: (value, force) ->
      return if !force and value == @filterValue # No actual change
      @filterValue = value
      if @filterValue == ''
        # No filter
        @collection.reset(@originalCollection.models)
      else
        @collection.reset(@originalCollection.filter((model) =>
          model.get('name').toLowerCase().indexOf(@filterValue) != -1
        ))

    updateSavingBanner: ->
      @$el.toggleClass('saving-toggle', !!@originalCollection.actions)
      @$el.removeClass('saving-error-toggle') if @originalCollection.actions
