require 'json'
require 'eloqua'
require 'signature'
require 'account_helper'
require 'recaptcha_helper'
require 'sign_up_validator'
require 'marketplace_sign_up'

class AccountsController < ApplicationController
  class ValidationError < StandardError; end

  include Eloqua
  include AccountHelper
  include ApplicationHelper
  include MarketplaceSignUp

  layout "outside"

  skip_before_action :check_for_password_reset
  skip_before_action :check_for_session_expiration
  skip_before_action :require_login

  before_action :redirect_to_sign_in_if_sign_up_is_disabled,
                :only => [:create, :create_psc, :new, :send_verification, :verify]
  before_action :initialize_errors,
                :only => [:create, :create_psc, :new, :send_verification, :verify]
  before_action :redirect_to_sign_in_if_marketplace_sentinel_is_expired,
                :only => [:create, :new]

  def new
    params[:user] ||= {}
  end

  def send_verification
    try_and_catch_validation_errors 'new' do
      validate_recaptcha if require_recaptcha?
      validate_chef_user
      expires = 1.day.from_now.to_i

      UserVerifyMailer.user_verify(
        params[:user],
        params[:ref],
        expires,
        Signature.new(
          params[:user][:username],
          expires,
          Chef::Manage::Application.config.secret_token
        )
      ).deliver_now
    end
  end

  def verify
    @expires = 1.day.from_now.to_i
    @user = valid_user?
    @ref = params[:ref]

    if @user.nil? || !valid_signature?
      flash[:alert] = I18n.t('errors.messages.invalid_signature')
      redirect_to :action => 'new'
      return
    end
  end

  def create
    if params[:user].nil? || !valid_user_signature?
      flash[:alert] = I18n.t('errors.messages.invalid_signature')
      redirect_to :action => 'new'
    else
      try_and_catch_validation_errors 'verify' do
        reset_session # Blank out the session just to be safe
        validate_chef_user
        if marketplace_sign_up_enabled?
          validate_marketplace_user
          register_marketplace_user
        end
        create_chef_user
        submit_user_to_eloqua if Settings.eloqua.enabled
        create_organization

        # Log the user in
        session[:username] = params[:user][:username]

        # Set initial value of password reset time (this will invalidate any sessions
        # for this user if any bogus sessions somehow otherwise exist)
        user_store = ChefManageUserStore.new(Settings.redis.url)
        user_store.initialize_password_reset(session[:username])

        session[:session_start_time] = Time.zone.now

        if community?
          render 'create_community'
        else
          redirect_to root_path
        end
      end
    end
  end

  # Accounts created for the AWS Product Support Connection (PSC) are intended
  # only to represent entitlement to support from Chef--not for actual login
  def create_psc
    try_and_catch_validation_errors 'verify' do
      reset_session # Blank out the session just to be safe
      validate_chef_user
      create_chef_user
      submit_user_to_eloqua if Settings.eloqua.enabled
      create_organization

      render plain: 'Created', content_type: 'text/plain', status: 201
    end
  end

  def validate_signup
    render SignUpValidator.new(server_root).validate(params)
  end

  private

  def initialize_errors
    @errors = []
  end

  def try_and_catch_validation_errors(template)
    begin
      yield
    rescue => e
      if e.is_a? ValidationError
        @errors ||= []
        @errors << e
        @user = params[:user]
        @ref = params[:ref]
        render template
      else
        raise
      end
    end
  end

  def redirect_to_sign_in_if_sign_up_is_disabled
    if Settings.disable_sign_up || external_auth_enabled?
      flash[:alert] = I18n.t('accounts.sign_up_disabled') if Settings.disable_sign_up
      redirect_to login_path and return
    end
  end


  def validate_recaptcha
    r = params[:'g-recaptcha-response']
    raise ValidationError.new "Missing reCAPTCHA validation." if r.blank?
    begin
      resp = RecaptchaHelper.verify(r, recaptcha_secret_key,
                                    Settings.recaptcha.verify_url)
    rescue StandardError => e
      logger.error("Failed reCAPTCHA validation: #{e}")
      # NOTE(ssd) 2017-05-30: We "fail-open" by default.
      # The primary use case of this feature is Hosted Chef where
      # we want to error on the side of continuing to allow sign-ups.
      if Settings.recaptcha.fail_open
        logger.warn("recaptcha.fail_open=true continuing with signup despite failure")
        resp = { "success" => true }
      else
        raise ValidationError.new "Failed reCAPTCHA validation."
      end
    end

    if !resp['success']
      logger.debug("Failed reCAPTCHA validation: #{resp}")
      raise ValidationError.new "Failed reCAPTCHA validation."
    end
  end


  def validate_chef_user
    raise ValidationError.new 'Missing form fields' if params[:user].blank?

    @validated_user = %w{username email display_name}.inject({}) do |memo, field|
      unless params[:user][field].present?
        raise ValidationError.new "Missing '#{field}'"
      end
      memo[field] = params[:user][field].to_s
      memo
    end

    @validated_user["password"] = params[:user][:password] if params[:user][:password]

    # Normalize between display_name and first/last
    @validated_user['display_name'] = @validated_user['display_name'].strip
    unless (@validated_user['display_name'].match(/^[^":*+=%?<>^$@{}()!_#&`~.|\/\\]+$/))
      raise ValidationError.new '^":*?<>^$|\/.@{}()!_#&`~=+% are not allowed in Full Name'
    end
    parts = @validated_user['display_name'].split(' ', 2)
    if parts.length == 2
      @validated_user['first_name'], @validated_user['last_name'] = parts
    else
      @validated_user['last_name'] = @validated_user['display_name']
    end

    params[:user][:first_name] = @validated_user['first_name']
    params[:user][:last_name] = @validated_user['last_name']
    params[:user][:display_name] = @validated_user['display_name']

    if terms_acceptance_required_on_sign_up? &&
       params[:user][:terms].blank?
      raise ValidationError.new 'Terms not accepted'
    end
  end

  def create_chef_user
    begin
      @private_key = server_root.post_rest('users', @validated_user)['private_key']
    rescue Net::HTTPServerException => e
      if [400, 409].include? e.response.code.to_i
        raise ValidationError.new parsed_json_for_error(e)['error']
      else
        raise
      end
    end
  end

  # Both creates the org and joins the user to it
  def create_organization
    return if params[:organization].blank?
    raise "Invalid slug: #{org_slug}" if org_slug.blank?
    username = params[:user][:username]
    # <ake the organization create call to opscode-account
    server_root.post_rest("organizations", {
      full_name: params[:organization].to_s,
      name: org_slug,
      org_type: 'Business',
    })

    # Make an association request for the current user to the new org
    association_request_response = server_root.post_rest("organizations/#{org_slug}/association_requests", user: username)
    association_request_id = association_request_response['uri'].split('/').last
    # Accept the association request as the user
    server_root.put_rest("users/#{username}/association_requests/#{association_request_id}", response: 'accept')

    # Add the user to the admins group and the billing admins group
    %w{admins billing-admins}.each do |groupname|
      group = server_root.get_rest("organizations/#{org_slug}/groups/#{groupname}")
      update_body_hash = {
        groupname: groupname,
        actors: {
          "users" => group["actors"].concat([username]),
          "groups" => group["groups"],
        }
      }
      server_root.put_rest("organizations/#{org_slug}/groups/#{groupname}", update_body_hash)
    end
  end

  def body_class
    'sessions'
  end

  def valid_user?
    begin
      input = decode_user_params
    rescue
      return nil
    end

    { 'username' => input[:username],
      'email' => input[:email],
      'display_name' => input[:display_name],
      'company' => input[:company],
      'terms' => input[:terms]
    }
  end

  def decode_user_params
    unsafe_input_encoded = params[:user]
    JSON.parse(Base64.urlsafe_decode64(unsafe_input_encoded),
                        symbolize_names: true,
                        create_additions: false
                      )
  end

  def valid_signature?
    params[:signature] &&
      params[:expires] &&
      @user['username'] &&
      Signature.new(
        @user['username'],
        params[:expires],
        Chef::Manage::Application.config.secret_token,
        decode_user_params.with_indifferent_access
      ).valid_for?(params[:signature])
  end

  def valid_user_signature?
    params[:button] &&
    Signature.new(
      params[:user][:username],
      params[:expires],
      Chef::Manage::Application.config.secret_token,
      filter_user_params.to_h
    ).valid_for?(params[:button])
  end

  def filter_user_params
    permitted_params[:user].select { |key, value| %w(username email display_name company terms).include?(key) }
  end

  def permitted_params
    params.permit!
  end

  def org_slug
    @org_slug ||= params[:organization].to_s.downcase.gsub(/[^\w\s-]/, '').strip.gsub(/\s+/, '-')
  end

  helper_method :org_slug
end
