require 'chef_rest_helpers'
require 'dark_launch'
require 'chef_manage_user_store'

class ApplicationController < ActionController::Base
  include ChefRestHelpers
  include PreloadHelpers
  protect_from_forgery
  before_action :set_locale
  before_action :check_for_password_reset
  before_action :check_for_session_expiration
  before_action :require_login
  before_action :assign_client_config_instance_var, unless: -> { request.format.json? }
  attr_reader :body_class

  unless Rails.application.config.consider_all_requests_local
    rescue_from ::Exception, :with => lambda { |exception| render_error 500, exception }
    rescue_from ::ActionController::RoutingError, AbstractController::ActionNotFound, ::ActionController::UnknownFormat, ::AbstractController::ActionNotFound, :with => lambda { |exception| render_error 404, exception }
  end

  # These errors are from Chef server
  rescue_from ::Net::HTTPServerException, :with => lambda { |exception|
    code = exception.response.code

    # Don't send forbidden and not found errors to Sentry, but do track
    # everything else.
    capture_exception(exception) unless [403, 404].include?(code)

    render_error code, exception
  }

  helper_method :body_class

  ACL_CONTAINERS = %w{clients groups containers data nodes roles cookbooks sandboxes environments}

  # If a port number is defined in the config, make sure we use that for URL
  # helpers.
  def default_url_options
    port = Settings.public_port
    port.nil? ? {} : { port: port }
  end

  def set_locale
    I18n.locale = params[:locale] || I18n.default_locale
  end

  def logout
    respond_to do |format|
      session[:username] = nil
      session[:referer] = request.original_url
      format.html { redirect_to login_url }
      format.json { head 401 }
    end
  end

  # Check to see if password has been reset for this user since this session was
  # created; if so, invalidate session and force user to log in with new password
  def check_for_password_reset
    if (logged_in)
      user_store = ChefManageUserStore.new(Settings.redis.url)
      if user_store.check_password_reset_time(session_username,
          session[:session_start_time])
        logout
      end
    end
  end

  # Check to see if this session has timed out
  def check_for_session_expiration
    # No point in doing anything unless user logged in
    if (logged_in)
      # timeout_limit of 0 means there's no limit, so skip checks
      if (timeout_limit != 0)
        if (session[:session_start_time] < Time.zone.now - timeout_limit)
          logout
        end
      end
    end
  end

  def logged_in
    # It's possible to have a username in the session and not be logged in (doing a
    # password reset, for instance), but the session_start_time is initialized
    # whenever actually logged in
    session_username.present? && session[:session_start_time]
  end

  def timeout_limit
    # Default is one week
    if (Settings.session_timeout_absolute.nil?)
      1.week
    else
      Settings.session_timeout_absolute.minutes
    end
  end

  def require_login
    unless session_username.present?
      session[:referer] = request.original_url
      respond_to do |format|
        format.html { redirect_to login_url }
        format.json { head 401 }
      end
    end
  end

  protected

  def assign_client_config_instance_var
    @client_config = { config: Settings }
  end

  # Return the current username, mostly here as a test hook
  def session_username
    session[:username]
  end

  # Override RestHelpers.server with the correct user info by default
  def server(user=nil, organization=nil)
    super(user || session_username, organization || params[:organization_id])
  end

  # Helper to get all the ACEs that a user matches for a given object
  # Pending backend support, so just a stub for now
  def fetch_matched_aces(object)
    {'read' => true, 'create' => true, 'grant' => true, 'update' => true, 'delete' => true}
  end

  def render_app
    # Load user data for prefab
    @user = begin
      server_root.get_rest("users/#{session_username}")
    rescue Net::HTTPServerException => e
      if %w{401 404 406}.include?(e.response.code)
        # Some kind of invalid user situation, to be safe lets void their session
        session[:username] = nil
        return redirect_to login_url
      else
        raise # Up the chain
      end
    end

    # if the user requested an id and we are on the org controller
    # we need to check that the org matching the id requested exists below
    if params.has_key?("id") && params['controller'] == 'organizations'
      found = false
    else
      found = true
    end

    # Load all of the user's orgs for prefab
    @organizations = orgs_for_current_user.map do |org_data|
      org = org_data['organization']
      # Whitelist just the keys we want because there are a lot of them
      org.delete_if { |key, value| !%w{name full_name clientname billing_plan}.include?(key) }
      org['access'] = ACL_CONTAINERS.inject({}) do |memo, container|
        ace = fetch_matched_aces("containers/#{container}")
        # Only CRG matter on containers
        ace.delete_if {|key, value| key != 'read' && key != 'create' && key != 'grant'}
        memo[container] = ace
        memo
      end

      # adds a hash darklaunch { dark_launch_field => true, field2... } to the main org hash
      org['darklaunch'] = DarkLaunch.all_features(org['name'])

      # hide "Reports" tab if reporting add on is not installed on chef server
      org['darklaunch']['reporting'] = reporting_addon_installed?

      # if the user requested a specific org and we found it, mark it so
      if params.has_key?("id") && params['controller'] == 'organizations' && org['name'] == params['id']
        found = true
      end

      org
    end

    if found
      # Add found user and orgs to client config
      @client_config.merge!(user: @user, organizations: @organizations)

      render 'dashboard/show'
    else #if the user requested a specific org and it was not found, 404
      render_error(404, ActionController::RoutingError.new('Organization does not exist'))
    end
  end

  def render_error(status, exception = nil)
    if exception
      capture_exception exception
      logger.error exception.message
      logger.error exception.backtrace.join("\n") if exception.backtrace.present?
    end
    @body_class = 'server-error'
    respond_to do |format|
      format.html do
        template = "errors/error_#{status}"
        template = 'errors/error_500' unless template_exists?(template)
        @status = status
        render template: template, layout: 'outside', status: status
      end
      format.json { render json: parsed_json_for_error(exception), status: status }
      format.all { render body:nil, :status => status }
    end
  end

  private

  def orgs_for_current_user
    @orgs_for_current_user ||= server_root.get_rest(
      "users/#{session_username}/organizations"
    ).sort_by do |org|
      org['organization']['name']
    end
  end

  def reporting_addon_installed?
    @installed ||= begin
      response = server_root.get_rest("reports/status")
      is_valid_response?(response)
    rescue Net::HTTPServerException
      false
    end
  end

  def is_valid_response?(response)
    response && response.has_key?("rest_api") && response["rest_api"] == "online"
  end

end
