require 'config_rb'
require 'starter_kit'
require "user_dissociation"

class OrganizationsController < ApplicationController
  # GET /type
  def index
    # Do nothing, just to support url navigation in backbone
    respond_to do |format|
      format.html { render_app }
    end
  end

  # GET /type/:id
  def show
    respond_to do |format|
      format.html { render_app }
    end
  end

  # GET /organizations/exists/:id
  #
  # different from SHOW in that it looks server-wide for org
  # and not just in /users/:id/organizations/:org_id
  def exists
    begin
      exists = server_root.get_rest("organizations/#{params[:id]}")
    rescue Net::HTTPServerException => e
      if e.response.code.to_i == 404
        render :json => {:success => true}, :status => 200 and return
      end
    end
    render :json => {:conflict => true, :organization => exists}, :status => 409
  end

  # GET /organizations/new
  def new
    respond_to do |format|
      format.html { render_app }
    end
  end

  # POST /organizations
  def create
    username = session[:username]
    org_slug = params[:organization][:name].to_s.downcase.gsub(/[^\w\s-]/, '').strip.gsub(/\s+/, '-')
    raise "Invalid slug: #{org_slug}" if org_slug.blank?

    # make the organization create call to opscode-account
    organization = server_root.post_rest("organizations", {
      full_name: params[:organization][:full_name],
      name: org_slug,
      org_type: params[:organization][:org_type]
    })

    # Make an association request for the current user to the new org
    association_request_response = server_root.post_rest("organizations/#{org_slug}/association_requests", user: username)
    association_request_id = association_request_response['uri'].split('/').last
    # Accept the association request as the user
    server_root.put_rest("users/#{username}/association_requests/#{association_request_id}", response: 'accept')

    # Add the user to the admins group and the billing admins group
    %w{admins billing-admins}.each do |groupname|
      group = server_root.get_rest("organizations/#{org_slug}/groups/#{groupname}")
      update_body_hash = {
        groupname: groupname,
        actors: {
          "users" => group["actors"].concat([username]),
          "groups" => group["groups"],
        }
      }
      server_root.put_rest("organizations/#{org_slug}/groups/#{groupname}", update_body_hash)
    end

    render json: organization, status: 201
  end

  # POST /organizations/<id>/leave
  def leave
    leave_org = UserDissociation.new(server_root, session_username, params[:id])

    # if the user leaving this org would make the admins group empty, render 403
    if leave_org.last_user_in_admins_group?
      render :json => {:success => false,
                       :message => "This user leaving the org would leave the admins group empty",
                       :failure => "admins"}, :status => 403

    # if the user leaving this org would make the billing-admins group empty, render 403
    elsif leave_org.last_user_in_billing_admins_group?
      render :json => {:success => false,
                       :message => "This user leaving the org would leave the billing-admins group empty",
                       :failure => "billing-admins"}, :status => 403
    else

      # delete user and render success
      server.delete_rest("users/#{session_username}")
      render :json => {:success => true}, :status => 200
    end
  end

  # GET /organizations/<id>/getting_started
  def getting_started
    respond_to do |format|
       format.html { render_app }
    end
  end

  # GET /organizations/<id>/knife_config
  def knife_config
    config_rb = ConfigRb.new(params[:id], session[:username])
    response.headers["Content-Disposition"] = config_rb.content_disposition
    send_data(config_rb, config_rb.send_data_options)
  end

  def get_starter_kit
    user_organizations = server_root.get_rest("users/#{session_username}/organizations")
    if user_organizations.length == 1
      redirect_to :controller => 'organizations',
      :action => 'starter_kit',
      :id => user_organizations[0]["organization"]["name"]
    else
      render_app
    end
  end

  # POST /organizations/<id>/getting_started
  def starter_kit
    # Get the user record now to ensure we're dealing with a valid user.
    user = server.get_rest("users/#{session[:username]}")

    # Reset the user key.
    user['private_key'] = true
    user.merge!(server_root.put_rest("users/#{session[:username]}", user))

    starter_kit = StarterKit.new(params[:id], user)
    response.headers['Content-Disposition'] = starter_kit.content_disposition
    send_data(starter_kit.to_zip, starter_kit.send_data_options)
  end

  protected

  # Override ApplicationController.server with the correct user info by default
  def server(user=nil, organization=nil)
    super(user || session_username, organization || params[:id])
  end

  private

  def user_has_no_orgs?(username)
    server_root.get_rest("users/#{username}/organizations").empty?
  end
end
