require 'signature'
require 'exception_helper'

class PasswordResetsController < ApplicationController
  include ApplicationHelper
  include AccountsHelper

  layout 'outside'

  before_action { redirect_to external_password_reset_url if use_external_password_reset? }
  before_action { redirect_to login_path if external_auth_enabled? }

  skip_before_action :check_for_password_reset
  skip_before_action :check_for_session_expiration
  skip_before_action :require_login

  def create
    if params[:username].blank?
      flash.now[:alert] = 'Please enter a username or email address'
      render 'new', status: :not_found
      return
    end

    begin
      # Mail password reset link
      username = URI::UrlParser.escape(resolve_user_from_username_or_email(params[:username]))

      user = server_root.get_rest("users/#{username}")
      expires = 1.day.from_now.to_i
      signature = Signature.new(username, expires,
                                Chef::Manage::Application.config.secret_token)
      PasswordResetMailer.password_reset(user, expires, signature).deliver_now

      # Reset password reset time and store the reset token (this will invalidate
      # any sessions for this user if logged into another browser somewhere).
      user_store = ChefManageUserStore.new(Settings.redis.url)
      user_store.password_reset(username, signature)
      flash.now[:notice] = "If '#{params[:username]}' username/email address exists, a password reset email has been sent."
      status = :ok
    rescue Net::HTTPServerException => e
      if e.response.code.to_i == 404
        flash.now[:alert] = "If '#{params[:username]}' username/email address exists, a password reset email has been sent."
        status = :not_found
      elsif e.response.code.to_i == 406
        flash.now[:alert] = "If '#{params[:username]}' username/email address exists, a password reset email has been sent."
        status = :forbidden
      else
        raise
      end
    rescue UserEmailNotFoundException => e
      flash.now[:alert] = e.message
      status = :not_found
    end

    render 'new', status: status
  end

  def new
  end

  def show
    unless valid_signature?
      flash[:alert] = invalid_signature_message
      redirect_to action: 'new'
      return
    end
  end

  def update
    if params[:password].blank?
      flash.now[:alert] = 'Password must not be blank'
      render 'show', status: :forbidden
      return
    elsif !valid_signature?
      flash.now[:alert] = invalid_signature_message
      render 'show', status: :forbidden
      return
    else
      begin
        user = server_root.get_rest("users/#{params[:username]}")
        server_root.put_rest("users/#{user['username']}",
                             user.merge('password' => params[:password]))
        session[:username] = user['username']
        session[:session_start_time] = Time.zone.now
        # Clear the password reset signature and login failure count once it's used
        user_store = ChefManageUserStore.new(Settings.redis.url)
        user_store.invalidate_password_reset_signature(user['username'])
        user_store.reset_login_failures(user['username'])
        redirect_to root_path
      rescue Net::HTTPServerException => e
        if e.response.code.to_i == 404
          flash[:alert] = "No user found for '#{params[:username]}'"
          redirect_to 'new'
          return
        elsif e.response.code.to_i == 400
          flash.now[:alert] = parsed_json_for_error(e)['error']
          render 'show', status: e.response.code
          return
        else
          raise
        end
      end
    end
  end

  private

  def invalid_signature_message
    'The given password reset link is expired or has an invalid signature'
  end

  def valid_signature?
    # Check that the params are present, the signature is valid and matches what we
    # have stored
    params[:signature].present? &&
      params[:username].present? &&
      params[:expires].present? &&
      ChefManageUserStore.new(Settings.redis.url).
          valid_password_reset_signature(params[:username], params[:signature]) &&
      Signature.new(params[:username], params[:expires],
          Chef::Manage::Application.config.secret_token).valid_for?(params[:signature])
  end
end
