require 'dark_launch'
require 'signature'
require 'account_helper'
require 'exception_helper'
require 'eloqua'

class SessionsController < ApplicationController
  include ApplicationHelper
  include AccountHelper
  include AccountsHelper
  include Eloqua

  layout "outside"

  skip_before_action :check_for_password_reset
  skip_before_action :check_for_session_expiration
  skip_before_action :require_login

  def new
    return redirect_oidc if saml_enabled?

    params[:user] ||= {}
    params[:to] = session[:referer]
    reset_session
  end

  def create
    return redirect_back if session[:username].present? # Already logged in
    return redirect_oidc if saml_enabled?

    unless auth = authenticate(params[:username], params[:password])
      # if we have a detailed error message, only show that one
      @login_fail = true unless @flash_error
      render 'new' and return
    end

    user, status = auth.values_at('user', 'status')

    # Force a reset to avoid session fixation attacks
    reset_session

    unless ["unlinked", "linked"].include?(status)
      raise "Unknown authentication status #{status.inspect}"
    end

    if status == 'unlinked'
      session[:link_account_authenticated_user] = user
      redirect_to account_links_path and return
    end

    session[:username] = user['username']

    # For security, we log session start time here
    session[:session_start_time] = Time.zone.now

    redirect_back
  end

  def destroy
    session[:username] = nil
    redirect_back
  end

  private

  def authenticate(username, password)
    return unless username.present? && password.present?

    # Resolve if username is email
    begin
      username = URI::UrlParser.escape(resolve_user_from_username_or_email(username))
    rescue UserEmailNotFoundException => e
      @flash_error = e.message
      return false
    end

    # Check for max login failures
    if (!external_auth_enabled?)
      user_store = ChefManageUserStore.new(Settings.redis.url)
      max_failures = Settings.max_login_failures
      if (max_failures && max_failures > 0 &&
          user_store.login_failure_count(username) >= max_failures)
        @login_attempts_failure = true
        return
      end
    end

    # Authenticate
    server_root.post_rest(
      'authenticate_user',
      :username => username, :password => password
    )
  rescue Net::HTTPServerException, Net::HTTPFatalError => e
    @ldap_failure = e.response.code == "504"
    raise unless @ldap_failure || e.response.code == "401"
    if (!external_auth_enabled? && e.response.code == "401")
      # If authentication fails, update login failures IF user exists
      begin
        # This will raise 404 (and be caught) if user doesn't exist, we don't want
        # to store/increment anything for non-existent users
        server_root.get_rest("users/#{username}")
        user_store.increment_login_failures(username)
        return
      rescue Net::HTTPServerException, Net::HTTPFatalError => e
        raise unless e.response.code == "404"
      end
    end
  end

  def redirect_oidc
    redirect_to oidc_login_path(to: redirect_destination(params.fetch(:to, '/')))
  end

  def redirect_back
    redirect_to redirect_destination(params.fetch(:to, '/'))
  end

  def body_class
    'sessions'
  end
end
