module FeedbackMailerHelper
  FormatValueDefaultOpts = { :wrap_tag => :pre, :indentation => 4 }
  YamlHeaderExpr = /\A---\s+/

  # Converts the given value to YAML if it's an Enumerable, as a naive way to
  # make the data we send with feedback emails human-readable. The output is
  # optionally wrapped in an HTML element (<tt><pre></tt> by default).
  # Non-Enumerable values will never be wrapped. For aesthetics the YAML header
  # ("<tt>---</tt>") is omitted.
  #
  # @param [Object] value The object to be formatted
  # @param [Hash] opts Options; all except +:wrap_tag+ will be passed to
  #   to +Object#to_yaml+
  # @option opts [Symbol] :wrap_tag (:pre) The HTML tag to wrap the output
  #   with, or +nil+ or +false+ to omit it
  # @option opts [Fixnum] :indentation (4) The number of spaces to use for
  #   YAML indentation
  #
  # @example
  #     val = { "foo" => { "bar" => 1 } }
  #
  #     format_value_for_email(val)
  #     # => <pre>
  #     #    foo:
  #     #        bar: 1
  #     #    </pre>
  #
  #     format_value_for_email(val, :wrap_tag => false, :indentation => 8)
  #     # => foo:
  #     #            bar: 1
  #
  def format_value_for_email(value, opts={})
    return value unless value.is_a?(Enumerable)
    opts = opts.reverse_merge(FormatValueDefaultOpts)

    wrap_tag = opts.delete(:wrap_tag)
    yaml = value.to_yaml(opts).sub(YamlHeaderExpr, '') # Strip YAML header ("---\n")

    wrap_tag ? content_tag(wrap_tag, yaml) : yaml
  end
end
