# Base class for other webui workers
require 'chef_rest_helpers'

# Stuff related to custom DSL extensions
module WorkerDSL
  def task_abstract(val=nil)
    @task_abstract = !!val unless val.nil?
    @task_abstract || false
  end

  def task_name(val=nil)
    if val
      @task_name = val.to_s
    else
      @task_name ||= name.underscore.gsub(/_workers?/, '')
    end
  end

  def task_label(val=nil)
    if val
      @task_label = val
    else
      @task_label || name
    end
  end
end

# Stuff related to the #server helper
module WorkerChefRestHelpers
  def self.included(klass)
    klass.extend(ClassMethods)
  end

  module ClassMethods
    # Pull off the first two arguments and store them specially so it isn't exposed to the underlying worker
    def store_identity(msg)
      msg['chef_rest_user'] = msg['args'].shift
      msg['chef_rest_organization'] = msg['args'].shift
    end
  end

  # Set the identity stored when the task was queued
  def set_identity(msg)
    @chef_rest_user = msg['chef_rest_user']
    @chef_rest_organization = msg['chef_rest_organization']
  end

  protected

  def server(user=nil, organization=nil)
    super(user || @chef_rest_user, organization || @chef_rest_organization)
  end
end

# Methods related to the life cycle tracking in Redis
module WorkerLifeCycle
  def self.included(klass)
    klass.extend(ClassMethods)
  end

  REDIS_OPTIONS = Settings.redis
  REDIS_POOL = ConnectionPool.new(:size => REDIS_OPTIONS.worker_pool_size, :timeout => REDIS_OPTIONS.worker_pool_timeout) do
    conn = ::Redis.new(:url => REDIS_OPTIONS.url, :driver => :hiredis)
    conn = ::Redis::Namespace.new(:actions, :redis => conn)
    conn
  end

  module ClassMethods
    def life_cycle_push(jid, org, data, expires=nil)
      REDIS_POOL.with do |redis|
        redis.multi do
          data[:updated_at] = Time.now.to_i.to_s
          jid_key = "jids:#{jid}"
          flattened_data = data.flatten.map do |val|
            if val.is_a?(Array) or val.is_a?(Hash)
              ActiveSupport::JSON.encode(val)
            else
              val
            end
          end
          redis.hmset(jid_key, *flattened_data)
          redis.expire(jid_key, expires) if expires
          org_key = "organizations:#{org}"
          redis.zadd(org_key, data[:updated_at], jid)
          data[:event_type] = 'action'
          data[:id] = jid # Even if not being updated in the hash, id must be sent on pubsub so the client knows what job this is
          data[:organization] = org # Ditto for org
          redis.publish(org_key, ActiveSupport::JSON.encode(data))
        end
      end
    end

    def life_cycle_queued(msg)
      data = {:id => msg['jid'], :label => task_label, :state => :queued, :args => msg['args'],
              :user => msg['chef_rest_user'], :organization => msg['chef_rest_organization']}
      life_cycle_push(msg['jid'], msg['chef_rest_organization'], data, 60*60*24) # Expire after one day
    end

    def for_organization(org, since=0)
      REDIS_POOL.with do |redis|
        org_key = "organizations:#{org}"
        jids = redis.zrangebyscore(org_key, since, '+inf')
        actions = redis.pipelined do
          jids.map {|jid| redis.hgetall("jids:#{jid}")}
        end
        empty_jids = jids.zip(actions).select{|jid, action| action.empty?}.map{|jid, action| jid}
        redis.zrem(org_key, empty_jids) unless empty_jids.empty?
        actions.reject{|action| action.empty?}
      end
    end
  end

  # Little wrapper around the class method version
  def life_cycle_push(data, expires=nil)
    self.class.life_cycle_push(jid, @chef_rest_organization, data, expires)
  end

  def life_cycle_started(msg)
    life_cycle_push({:state => :started})
  end

  def life_cycle_finished(msg)
    life_cycle_push({:state => :finished})
  end

  def life_cycle_error(msg, ex)
    # Keep errors for a week
    life_cycle_push({:state => :error, :error => ex.to_s}, 60*60*24*7)
  end

  def progress(n, total=nil)
    data = {:progress => n, :organization => @chef_rest_organization, :id => jid}
    data[:progress_total] = total if total
    life_cycle_push(data)
  end

  def expire(object_type, object_id)
    REDIS_POOL.with do |redis|
      data = {:event_type => 'expire', :organization => @chef_rest_organization, :object_type => object_type, :object_id => object_id, :id => jid}
      redis.publish("organizations:#{@chef_rest_organization}", ActiveSupport::JSON.encode(data))
    end
  end


  def delete_item(object_type, object_id, object_parent_type, object_parent_id)
    REDIS_POOL.with do |redis|
      data = {:event_type => 'delete_item', :object_type => object_type, :object_id => object_id, :id => jid, :parent_id => object_parent_id, :parent_type => object_parent_type, :organization => @chef_rest_organization}
      redis.publish("organizations:#{@chef_rest_organization}", ActiveSupport::JSON.encode(data))
    end
  end

  def update(object_type, object_id, operation, result, on_user_channel= false)
    REDIS_POOL.with do |redis|
      data = {:event_type => 'actions_result', :operation => operation, :organization => @chef_rest_organization, :object_type => object_type, :object_id => object_id, :result => result, :id => jid}
      if on_user_channel
        redis.publish("users:#{@chef_rest_user}", ActiveSupport::JSON.encode(data))
      else
        redis.publish("organizations:#{@chef_rest_organization}", ActiveSupport::JSON.encode(data))
      end
    end
  end

  def update_user(object_type, object_id, operation, result)
    update(object_type, object_id, operation, result, true)
  end

  def update_organization(object_type, object_id, operation, result)
    update(object_type, object_id, operation, result, false)
  end

end
class BaseWorker
  include Sidekiq::Worker
  extend WorkerDSL
  include ChefRestHelpers
  include WorkerChefRestHelpers
  include WorkerLifeCycle

  # This can be removed for Sidekiq > 2.5.2
  attr_accessor :jid

  class << self
    attr_accessor :tasks, :task_map

    # Find a worker class by name
    def by_name(name)
      raise "No worker classes found, check why workers/*.rb isn't loading" unless @tasks
      BaseWorker.task_map ||= begin
        BaseWorker.tasks.inject({}) do |memo, klass|
          memo[klass.task_name] = klass unless klass.task_abstract
          memo
        end
      end
      BaseWorker.task_map[name]
    end

    def inherited(klass)
      BaseWorker.tasks ||= []
      BaseWorker.tasks << klass
    end
  end

end
