(ns reagent-manage.ajax
  (:require
   [ajax.core]
   [reagent-manage.env :refer [env]]))

;; True iff we should mock ajax calls
(def mocking? (= env :test))

(defn make-expectation
  "Convert a http verb keyword and arguments into an expected call form."
  [kwd [url _]]
  [(name kwd) url])

;; An atom containing either nil (no ajax call expected) or a pair
;; consisting of an expected call form and a fake result to return to
;; the caller.
(def expected (atom nil))

(defn expect
  "Set the expectation state to the given ajax call."
  [kwd args result]
  (reset! expected [(make-expectation kwd args) result]))

(defn process-event
  "Given an ajax call event, reconcile it against the current mocking
  state. For now, we can only remember a single expected call, and so
  we just make sure the actual call corresponds to it, (throwing an
  exception if not) and we reset the expectation state at the end if
  it succeeds."
  [kwd args]
  (let [actual-event (make-expectation kwd args)]
    (if-let [[expected-event response] @expected]
      (if (= expected-event actual-event)
        (do
          ((:handler (second args)) response)
          (reset! expected nil))
        (throw (js/Error. (str "expected ajax call " expected-event
                               " does not match actual event " actual-event))))
      (throw (js/Error. (str "unexpected ajax call " actual-event))))))

(defn kwd->method
  "Map http verb keyword into real implementation function."
  [method]
  (case method
    :GET ajax.core/GET
    :POST ajax.core/POST
    :PUT ajax.core/PUT
    :DELETE ajax.core/DELETE))

(defn maybe-mock
  "Given an http verb keyword and some arguments, either really call
  the ajax function or else divert to ajax mocks."
  [kwd args]
  (if mocking?
    (process-event kwd args)
    (apply (kwd->method kwd) args)))

(defn GET [& args] (maybe-mock :GET args))
(defn POST [& args] (maybe-mock :POST args))
(defn PUT [& args] (maybe-mock :PUT args))
(defn DELETE [& args] (maybe-mock :DELETE args))

(defn with-mocks
  "Run supplied function with ajax mocking."
  [k]
  (reset! expected nil)
  (k))
