(ns reagent-manage.api
  (:require
   [reagent-manage.utils :as u]
   [reagent-manage.error :as error]
   [reagent-manage.env :refer [env]]
   [reagent.session :as session]
   [cljs.core.async :refer [chan put! close!]])
  (:require-macros
   [cljs.core.async.macros :refer [go]]))

;; Utility functions
;; -------------------------
;;

(defn search-indexes
  "Get list of all search indexes (e.g. nodes, clients, environments)
  from data provided by rails controller. This is in a very unadorned
  top-level place in the global data since what the Chef API provides
  at the route `/organizations/:organization_id/search` is exactly the
  list of search indexes and we're imitating its structure here.

  Exclude saved-searches since it's more likely to be confusing than
  useful in the interface, e.g. showing the index of saved-searches
  among the saved searches *of* various indexes."
  [] (dissoc (js->clj (.-data js/chef)) "saved-searches"))

(defn base-url
  "Base URL for all api calls. The rails controller takes care of
  supplying the appropriate `chef.organization_id` to us."
  []
  (str "/organizations/" (.. js/chef -organization_id)))

(if (not= env :test)
 (defn csrf-token
   "The CSRF token is necessary for Chef Infra Server request authentication
  to succeed."
   []
   (.. js/document
       (querySelector "meta[name='csrf-token']")
       (getAttribute "content")))
 (defn csrf-token [] "TEST-CSRF-TOKEN") )

(if (not= env :test)
 (defn animate-logo
   "Adjust the Chef logo's class list in response to an API call
  beginning (`doit = 1`) or completing (`doit = 0`)."
   [do-it]
   (let [logo (.item (.getElementsByClassName js/document "logo") 0)
         class-list (.-classList logo)
         n (if do-it 1 -1)
         new-count (:ajax-requests (session/update-in! [:ajax-requests] + n))]
     (if (> new-count 0)
       (.add class-list "animate")
       (.remove class-list "animate"))))
 (defn animate-logo [arg]))

;; API functions
;; -------------------------

(defn error-handler
  "Handle errors arising from Chef API calls."
  [{:keys [status status-text] :as resp}]
  (.log js/console (str "something bad happened: " status " " status-text))
  (animate-logo false)
  (error/report ["The Chef Infra Server reported an error: " [:br]
                 [:tt (str status " " status-text) " "
                  (get-in resp [:response :error 0])]]))

(defn convert-to-api-uri
  "Convert a uri arising from the Chef Infra Server (which we can't use
  directly for lack of CORS support) to a proxy path local to the
  rails server."
  [chef-server-uri]
  (clojure.string/replace chef-server-uri #"https://.*?/" "/api/"))

(defn request-with-handler
  "Make a Chef Infra Server API request. `method` is some verb (e.g. `GET`)
  from `ajax.core`. `path` is an absolute url. `options` are passed
  along to the `ajax.core` function. `handler` is a callback which
  receives the data on success."
  [method path handler options]
  (do
    (animate-logo true)
    (method path
            (merge {:headers {"X-CSRF-Token" (csrf-token)}
                    :handler (fn [resp]
                               (do
                                 (animate-logo false)
                                 (error/report nil)
                                 (handler resp)))
                    :format :json
                    :response-format :json
                    :keywords? true
                    :error-handler error-handler}
                   options))))

(defn request-absolute
  "Make a Chef Infra Server API request. `method` is some verb (e.g. `GET`)
  from `ajax.core`. `path` is an absolute url. `options` are passed
  along to the `ajax.core` function, returning a channel to which
  either the result or a js Error object is sent.

  Plausible reasons so far to use this include
  (a) you're getting a chef infra server uri from somewhere else and
  converting it with `convert-to-api-uri`. Most api calls want to go
  through plain old `request` below.
  (b) you're hitting an endpoint on the chef rails server rather than
  the chef-server api proper. Consider very carefully if you really
  have to do this! We're trying to confine the front-end to the
  chef-server api as much as possible."
  ([method path] (request-absolute method path {}))
  ([method path options]
   (let [c (chan)]
     (request-with-handler
      method path
      #(go
         (>! c %)
         (close! c))
      (assoc options :error-handler
             #(go
                (>! c (u/make-err %))
                (close! c))))
     c)))

(defn request
  "Make a Chef Infra Server API request. `method` is some verb (e.g. `GET`)
  from `ajax.core`. `path` is relative to an organization prefix.
  `options` are passed along to the `ajax.core` function, returning a
  channel to which either the result or a js Error object is sent."
  ([method path] (request method path {}))
  ([method path options] (request-absolute method (str "/api" (base-url) path) options)))
