(ns reagent-manage.core
  (:require
   [cljs.core.match :refer-macros [match]]
   [reagent-manage.api :as api]
   [reagent-manage.entity :as entity]
   [reagent-manage.error :as error]
   [reagent-manage.feedback :as feedback]
   [reagent-manage.fragment :as fragment]
   [reagent-manage.history :as history]
   [reagent-manage.search-interface :as search]
   [reagent-manage.utils :as u]
   [reagent.core :as reagent :refer [atom cursor]]
   [reagent.session :as session]))

;; Views
;; -------------------------

(defn sidebar
  "Sidebar component containing saved searches."
  [satm]
  (let [state @satm
        searches-by-index
        (merge
         (u/make-map (for [[name _] (api/search-indexes)] [name []]))
         (group-by (fn [[kwd uri]]  (get-in state [:search-data (name kwd) :index]))
                   (:saved-searches state)))
        sorted-searches (sort (u/comparef first) searches-by-index)
        sidebar-rows (apply concat (for [[index searches] sorted-searches]
                                     `[[:index ~index]
                                       ~@(for [search searches] [:search search])]))]
    [:div.large-3.medium-4.columns
     [:nav [:ul.side-nav
            [:li.heading [:a {:on-click #(search/load-saved-searches satm)}
                          [:i.fi-refresh] " Saved Searches"]]

            (for [row sidebar-rows]
              (match [row]

                     [[:index index]]
                     ^{:key index}
                     [:li.heading
                      [:a {:on-click #(search/new-search-with-index index satm)} index]]

                     [[:search [kwd uri]]]
                     (let [search-name (name kwd)]
                       ^{:key search-name}
                       [:li {:style {:margin-left "2em"}}
                        [:a {:on-click
                             (fn [e]
                               (search/set-current-search! satm search-name)
                               (search/get-saved-search-into-atom
                                uri
                                (cursor satm [:search-data search-name])))}
                         search-name]])))]]]))

(defn index-page
  "The main page containing the sidebar and search interface."
  [satm]
  (let [state @satm]
    [:div
     [fragment/generator (match
                          [(:current-entity state)]
                          [[:saved-search :new]] (history/index-route (get-in state [:search-data :new]))
                          [[:saved-search x]] (history/saved-search-route {:name x})
                          [[:node node-url]] (history/view-node-route {:url node-url}))]
     (if (:modal state) [(:modal state) satm] [:span])
     [error/alert satm]
     [sidebar satm]
     [:div.large-9.medium-8.columns
      (match [(:current-entity state)]
             [[:saved-search current-search]]
             (let [method (match current-search
                                 :new "POST"
                                 _    "PUT")]
               [search/form satm method (cursor satm [:search-data current-search])])
             [[:node url]]
             [entity/view url (cursor satm [:entity-view])])]]))

(defn mount-root-inner "Initialization function that connects reagent
  components to dom nodes."
  []
  (search/load-saved-searches session/state)

  ;; We want to have for the most part all interface state descended
  ;; from session/state. The reason that we're picking session/state
  ;; rather than doing like (let [state atom {...}]) on our own right
  ;; here is so that the state automatically persists across figwheel
  ;; code-reloads, which tends to be convenient for development.
  (reagent/render [index-page session/state] (.getElementById js/document "search-app"))
  (reagent/render [feedback/feedback-button session/state] (.getElementById js/document "feedback-button")))

(defn mount-root
  "The wrapper (e.g. `env/dev/cljs/reagent_manage/dev.cljs`) calls
  `reagent-manage.mount-root` as its `:jsload-callback`, but it captures
  the original value of `mount-root`, so changes to the function
  `mount-root` itself are not observed without reload. The trivial
  workaround is to just make it call another function. Since the
  *call-site* of mount-root-inner is in `core.cljs`, it's updated when
  `core.cljs` is edited."
  []
  (mount-root-inner))

(defn init!
  "Entry point coming from from `env/dev/cljs/reagent_manage/dev.cljs`
  or `env/prod/cljs/reagent_manage/dev.cljs`."
  []
  (history/init!)
  (mount-root))
