(ns reagent-manage.feedback
  (:require [reagent.core :as reagent :refer [atom cursor]]
            [reagent.session :as session]
            [reagent-manage.utils :as u]
            [reagent-manage.modal :as modal]
            [reagent-manage.api :as api]
            [ajax.core :refer [POST]])
  (:require-macros
   [cljs.core.async.macros :refer [go]]
   [reagent-manage.macros :refer [go-api go-re <?]]))

(defn feedback-success-modal
  "Component to show after 'give feedback' is successful. "
  [satm]
  (let [close-fn #(swap! satm assoc :modal nil)]
    [modal/modal close-fn
     [:span "Feedback was sent successfully."]
     :bg-dismiss true :key-dismiss true]))

(defn no-feedback-modal-body
  "Component to render body of 'give feedback' modal dialog in the
  event that we don't support sending it, e.g. when in on-premises
  private chef mode."
  [modal-atm report-fn]
  (let [support-email (u/get-config "support_email_address")]
    (fn []
     [:span
      "To submit general feedback about the Management Console, please contact "
      [:a {:href (str "mailto:" support-email "?subject=Chef Management Console Feedback")}
       support-email]
      [:p "If this is an urgent support request, please log into your "
       "support ticketing system, create an issue, and mark it as High Severity."]])))

(defn feedback-modal-body
  "Component to render body of 'give feedback' modal dialog."
  [modal-atm report-fn]
  (let [subject-atm (atom "")
        desc-atm (atom "")]
    (fn []
      [:span
       [:h2 {:style {:margin-top "-0.3em" :margin-bottom "0.5em"}} "Feedback" ]
       [:div.row.collapse
        [:div.medium-2.columns [:label.inline.right "Subject"]]
        [:div.medium-10.columns [u/text-input subject-atm]]]
       [:div.row.collapse
        [:div.medium-2.columns [:label.inline.right "Description"]]
        [:div.medium-10.columns [u/textarea-input desc-atm {:rows 10}]]]
       [:span.right
        [:button.button.secondary
         {:on-click #(reset! modal-atm nil)}
         "Cancel"] " "
        [:button.button
         {:on-click (fn []
                      (go-api
                       (try
                         (<? (report-fn @subject-atm @desc-atm))
                         (reset! modal-atm #'feedback-success-modal)
                         (catch js/Object e
                           (reset! modal-atm nil)
                           (throw e)))))}
         "Send Feedback"]]])))

(defn get-browser-info
  "Returns some browser information, based on `getBrowserInfo` in
  `chef-manage/app/assets/javascripts/views/controls/modals/report_feedback.coffee`"
  []
  {:navigator
   (-> (js* "function() {rv={};for(k in navigator){rv[k]=navigator[k];}return rv;}")
       (#(%)) ;; call the above function
       (js->clj)
       (dissoc "plugins" "mimeTypes"))
   :window_innerHeight js/innerHeight
   :window_innerWidth js/innerWidth})

(defn get-app-state
  "Returns some application state for feedback, based on `getAppState` in
  `chef-manage/app/assets/javascripts/views/controls/modals/report_feedback.coffee`"
  []
  {:session-state @session/state
   :chef-data (js->clj js/chef)
   :url (.-href js/location)})

(defn send-feedback
  "Send feedback with subject `subject` and body `desc` to the rails
  server for it to send via email."
  [subject desc]
  (go-re
   (<? (api/request-absolute
        POST
        (str (api/base-url) "/actions")
        {:params
         {:action_id "util/report_feedback"
          :arguments
          (.stringify js/JSON (clj->js [nil subject desc
                                        (get-browser-info)
                                        (get-app-state)]))}}))))

(defn feedback-modal
  "Component to show 'give feedback' modal dialog. "
  [satm]
  (let [close-fn #(swap! satm assoc :modal nil)]
    [modal/modal close-fn
     (if (u/get-config "feedback_form_enabled")
       [feedback-modal-body (cursor satm [:modal]) send-feedback]
       [no-feedback-modal-body])]))

(defn feedback-button
  "Component to show 'report feedback' button."
  [satm]
  [:a {:href "#"
       :on-click
       (fn [e]
         (swap! satm assoc :modal #'feedback-modal)
         (.preventDefault e))}
   [:i.fi-comment] " Report Feedback"])
