(ns reagent-manage.history
  (:require
   [clojure.set]
   [goog.events :as events]
   [goog.history.EventType :as EventType]
   [reagent-manage.entity :as entity]
   [reagent-manage.panels :as panels]
   [reagent-manage.search-interface :as search]
   [reagent.session :as session]
   [secretary.core :as secretary :include-macros true])
  (:import goog.history.Html5History))

(defn initial-state
  "Initial application state, stored in the atom `reagent.session/state`."
  []
  (let [basic-state
        {;; If a modal dialog box is being shown, has a value which is a symbol
         ;; of the component to render it. Otherwise nil.
         :modal nil

         ;; Just the raw json response from the chef-server listing which
         ;; saved-searches data items exist.
         :saved-searches {}

         ;; Cached data about searches and search results.
         :search-data {:new search/initial-search}

         ;; Can be:
         ;; [:saved-search ss] where ss is a key in the map in :search-data
         ;; [:node url] where url is a chef-server api url for the node
         :current-entity [:saved-search :new]

         ;; Number of ajax-requests outstanding so we can show progress.
         :ajax-requests 0

         ;; Whether to show various optional panels,
         :visible-panels #{
                           ;; not :advanced-search
                           ;; not :saved-search
                           :welcome-message
                           }

         ;; Non-nil when there should be an error displayed in an alert.
         :error nil}]
    (update basic-state :visible-panels #(clojure.set/difference % (panels/perma-hidden)))))

;; Routes
;; -------------------------

(secretary/set-config! :prefix "#")

(secretary/defroute root-route
  "/" []
  (swap! session/state (fn [state] (merge state (dissoc (initial-state) :saved-searches)))))

(secretary/defroute saved-search-route
  "/saved/:name" [name]
  (search/set-current-search! session/state name))

(secretary/defroute index-route
  "/new/:index" [index]
  (search/new-search-with-index index session/state))

(secretary/defroute view-node-route
  "/node*url" [url]
  (entity/set-current-entity! session/state [:node url]))

;; History
;; -------------------------

;; (must be called after routes have been defined)
(defn history-handler [event]
  (.log js/console event)
  (let [fragment (.-token event)]
    (if (= fragment "")
      (print "empty fragment")
      (secretary/dispatch! (.-token event)))))

(defn hook-browser-navigation!
  "Connect to google closure library history management. Not being
  used currently until we can figure out how do work out the bugs in
  it, see below (\"What used to be here\")."
  []
  (let [history (Html5History.)]
   (doto history
     (events/listen EventType/NAVIGATE history-handler)
     (.setEnabled true))))

(defn init!
  "Initialize state and history tracking."
  []
  (reset! session/state (initial-state))
  (let [hash (.-hash js/location)]
    (if (= hash "")
      (let [default-state (index-route {:index "node"})] (.replaceState js/history default-state "" default-state))
      (secretary/dispatch! hash)))
  (set! (.-onpopstate js/window) #(secretary/dispatch! (.-state %))))
