(ns reagent-manage.json-browser
  (:require
   [reagent.core :as reagent :refer [atom cursor]]
   [reagent.session :as session])
  (:require-macros
   [reagent.ratom :refer [reaction]]))

(defn typeof
  "Like javascript typeof (i.e. return the type of `x`) except we also
  try to consistently detect arrays and null values."
  [x]
  (if x
   (let [raw-type ((js* "function(x){return typeof(x);}") x)]
     (if (and (= raw-type "object") (= (.-constructor x) js/Array))
       "array"
       raw-type))
   "null"))

(defn show-string
  "Display a human-readable form of `x`. Doesn't presently do any
  escaping, but probably should."
  [x]
  (str "\"" x "\""))

(defn bool?
  "Return true iff the argument is `true` or `false`. Not sure why I
  can't find this as a standard library function."
  [x]
  (case x
    true true
    false true
    false))

(defn show-clj
  "Display the clojurescript data in `clj`, indented by the string
  `indent` which consists of some spaces, but display only a truncated
  version if `revealed-cur` derefs to false. In general, the top-level
  atom corresponding to `revealed-cur` contains a hierarchy of maps
  that expand out as far as we want data to be shown. "
  [data indent revealed-cur]
  (if @revealed-cur
    `[:span
      ~@(cond
          (nil? data) "nil"
          (map? data) ["{"
                       (for [key (keys data)]
                         ^{:key key} [:span "\n " indent (pr-str key) " "
                                      [show-clj (get data key) (str " " indent)
                                       (cursor revealed-cur [key])]
                                      ])
                       "\n" indent "}"]
          (vector? data) ["["
                          (for [ix (range (count data))]
                            ^{:key ix} [:span "\n " indent
                                        [show-clj (get data ix) (str " " indent)
                                         (cursor revealed-cur [ix])]
                                        ])
                          "\n" indent "]"]
          true ["#<Expanded but Unexpandable>"])]
    (let [expander [:a {:on-click (fn [e] (reset! revealed-cur {}) (.preventDefault e))} "..."]]
      `[:span ~@(cond
                  (nil? data) "null"
                  (map? data) ["{ " expander " }"]
                  (vector? data) ["[ " expander " ]"]
                  (string? data) [(show-string data)]
                  (number? data) [data]
                  (bool? data) [(pr-str data)]
                  true [(str "#<Unknown: " (pr-str data) ">")])])))

(defn json-browser
  "Component to display the javascript data in `json`."
  [json]
  (let [revealed-atm (atom {})]
    (fn [json]
      [:pre [show-clj (js->clj json) "" revealed-atm]])))

(defn clj-data-browser
  "Component to display the clojurescript data in `clj`."
  [clj]
  (let [revealed-atm (atom {})]
    (fn [clj]
      [:pre [show-clj clj "" revealed-atm]])))

;; For all of these open-in-a-new-window functions, be sure to enable
;; popups on localhost or you won't see anything!

(defn atm-browser [browser atm]
  [browser @atm])

(defn browser-in-new-window
  "Open a new window and mount the component `browser` on the data in
  atom `atm`."
  [browser atm]
  (print atm)
  (let [win (.open js/window "" "_blank" "other,width=400,height=300")]
    (.write (.-document win) "<div id=\"attach\"></div>")
    (reagent/render [atm-browser browser atm]
                    (.getElementById (.-document win) "attach"))))

(defn state-browser-in-new-window
  "Open a new window showing the current session state."
  []
  (browser-in-new-window clj-data-browser session/state))
