(ns reagent-manage.query-builder
  (:refer-clojure :exclude [update-in])
  (:require
   [reagent-manage.debug :refer [clog jsclog as-json]]
   [reagent-manage.utils :as u]
   [reagent.core :as reagent]
   [reagent.session :as session]
   [goog.events :as events]
   [clojure.string :refer [join]]
   [cljs.core.match :refer-macros [match]]
   [cljsjs.peg]))

;; The default library behavior of `(update-in m [] f)` is
;; broken, this fixes it.
(defn update-in
  ([m [& ks] f] (if ks (cljs.core/update-in m ks f) (f m))))

;; This deftype is a simplification of reagent's `RCursor`, plus the
;; ability to call a callback whenever a write happens. It supports
;; the `ILookup` protocol so that we can just use things like `get-in` to
;; effectively get write access down into the guts of a hierarchical
;; data structure.
;;
;; Possibly this can be done better with the tools already in
;; `reagent.ratom`.
(deftype WriteListener [ratom write-cb path]
  IDeref
  (-deref [this] (get-in @ratom path))

  ILookup
  (-lookup [this k default] (-lookup this k))
  (-lookup [this k] (WriteListener. ratom write-cb (concat path [k])))

  IReset
  (-reset! [a x]
    (-swap! ratom assoc-in path x)
    (write-cb))

  ISwap
  (-swap! [a f]
    (do (-swap! ratom #(update-in % path f))
        (write-cb))))

(defn write-listener
  "Construct a `WriteListener`"
  [ratom write] (WriteListener. ratom write []))


;; Query builder
;; -------------------------
;;
(defn with-default-conjoiner
  "Provide a default conjoiner"
  [conjoiner]
  (or conjoiner "AND"))

(defn search-builder-field
  "Component to render one field in the search builder. `element` is a
  hiccup template keyword indicating what kind of dom element it
  should be, `cell` is the field data."
  [element cell]
  (vec
   (concat
    [element]
    (case (:type @cell)
      "FIELD" [[u/text-input (get cell :attr)]
               [:span {:style {:font-size 24}} " = "]
               [u/text-input (get cell :val)]]
      "INC_RANGE"
      [[u/text-input (get cell :attr)]
       [:span {:style {:font-size 24}} " in [ "]
       [u/text-input (get cell :from)]
       [:span {:style {:font-size 24}} ", "]
       [u/text-input (get cell :upto)]
       [:span {:style {:font-size 24}} " ]"]]
      "EXC_RANGE"
      [[u/text-input (get cell :attr)]
       [:span {:style {:font-size 24}} " in ( "]
       [u/text-input (get cell :from)]
       [:span {:style {:font-size 24}} ", "]
       [u/text-input (get cell :upto)]
       [:span {:style {:font-size 24}} " )"]])
    )))

(defn search-builder-row
  "Component to render one entire row in the search builder. `search`
  is the entire search data structure, `i` is the index of the row."
  [search i]
  (let [conjoiner-field (u/wrap-reader (get-in search [i :conjoiner]) with-default-conjoiner)]
    [:tr
     ;; Remove-row button
     [:td [:button.button.delete.tiny {:on-click (fn [e] (swap! search #(vec (u/without-nth % i))))} u/times]]

     [:td (if (= i 0) ""
              [u/select-input conjoiner-field
               {:style {:width "5em"}}
               (u/select-option "AND" "and")
               (u/select-option "OR" "or")])]

     [:td
      [u/select-input (get-in search [i :type])
       {:style {:width "5em"}}
       (u/select-option "FIELD" "field")
       (u/select-option "INC_RANGE" "range")
       (u/select-option "EXC_RANGE" "exclusive range")]]
    [search-builder-field :td (get search i)]
     ]))

(defn search-builder-component
  "Component to render the entire search builder. `search` is the data
  structure representing the query. `close-fn` is a callback for when
  the close button is pressed."
  [search close-fn]
  (let [new-row {:attr ""
                 :val ""
                 :conjoiner "AND"
                 :type "FIELD"
                 }]
    [:div.search-builder.closable-container
     [:table
      [:tbody (map-indexed
               (fn [i row] ^{:key i} [search-builder-row search i]) @search)
       ;; Add-new-row button
       [:tr [:td [:button.button.tiny {:on-click (fn [e] (swap! search #(vec (concat % [new-row]))))} "+"]]]]]
     [:a.close {:on-click close-fn} u/times]]))

(defn default-star
  "If a field is left empty, its meaning defaults to \"*\"."
  [s] (if (or (= s "") (not s)) "*" s))

(defn search-row-stringify
  "Convert the row `row` of the query (which occurred at index `i`) to a string."
  [i row]
  (concat (if (= i 0) "" (str " " (with-default-conjoiner (:conjoiner row)) " "))
          (case (:type row)
            "FIELD" (join ":" [(default-star (:attr row))
                               (default-star (:val row))])
            "INC_RANGE" (str (default-star (:attr row))
                         ":[" (default-star (:from row)) " TO " (default-star (:upto row)) "]")
            "EXC_RANGE" (str (default-star (:attr row))
                         ":{" (default-star (:from row)) " TO " (default-star (:upto row)) "}"))))

(def grammar
  "PEG.js grammar for search queries."
  "

query = space? first:expression rest:conjunct* space? { return [first].concat(rest) }
      / space? {return []}

expression = field / field_range

field = name:$field_name ':' arg:$term { return {attr:name, val:arg, type:'FIELD'} }
field_range = name:$field_name ':' range:(incl_range / excl_range) {
 range.attr = name;
 return range;
}

range_entry = '*' / $valid_letter;

incl_range = '[' from:range_entry ' TO ' upto:range_entry ']' { return {type:'INC_RANGE',from:from,upto:upto} }
excl_range = '{' from:range_entry ' TO ' upto:range_entry '}' { return {type:'EXC_RANGE',from:from,upto:upto} }

conjunct = space conjoiner:$conjoiner space expression:expression {
 expression.conjoiner = conjoiner;
 return expression;
}

term = valid_letter

space = [ \\t]+

conjoiner = 'AND' / 'OR'

field_name = valid_letter

valid_letter = start_letter+ ([a-zA-Z0-9*?@_+./-] / '\\\\' special_char)*

start_letter = [a-zA-Z0-9_.*/] / '\\\\' special_char

special_char = '[' / ']' / '\\\\' / \"\\\"\" / [!(){}^~*?:]

")

(def parser
  "Precompute the parser from the grammar."
  (.buildParser js/PEG grammar))

(defn parse-string
  "Parse a query string, yielding `nil` if there are parse errors."
  [str]
  (try (js->clj (.parse parser str) :keywordize-keys true)
       (catch (.-SyntaxError parser) _ nil)))

(defn resolve-query-struct
  "Given that the current search data structure is `query-struct`, and the current state
  of the raw query is `raw-query`, returns `[query, in-harmony]`, answering two questions at once:

  * `query`: What should the user see right now as the text of the query?
  * `in-harmony`: Is this query text consistent with the search data structure?

The latter answer is 'no' when there is a parse error in the raw
  query, and is automatically 'yes' when `raw-query` is `nil` because
  that means we're letting the query data structure be entirely in
  charge of determining the user-visible query string."
  [query-struct raw-query]
  (let [computed-query
        (apply str (apply concat (map-indexed search-row-stringify query-struct)))]
    (if-let [val raw-query]
          [val (= query-struct (parse-string val))]
          [computed-query true])))

(defn get-doc-query
  "Given that the current search data is `doc`,
  returns what the user should see right now as the text of the
  query."
  [doc]
  (let [[query _] (resolve-query-struct (:query-struct doc) (:raw-query doc))]
    query))

(defn search-query-display
  "Component to render an input text box containing the user-visible query
  string. `struct-atm` is an atom containing the current search data structure,
  `raw-atm` is an atom containing the raw query string."
  [struct-atm raw-atm]
  (let [[displayed-query in-harmony] (resolve-query-struct @struct-atm @raw-atm)]
    [:input {:type "text"
             :style {:background-color (if in-harmony nil "#fee")}
             :value displayed-query
             :on-change (fn [e]
                          (let [newval (-> e .-target .-value)]
                            ;; In any event we're going to preserve the edit in atom `raw-atm`
                            (reset! raw-atm newval)
                            ;; ...and update the query data structure in
                            ;; `struct-atm` if parsing succeeds:
                            (if-some [parsed (parse-string newval)]
                              (reset! struct-atm parsed))))}]))

(defn refine-search
  "Given a query structure `query-struct`, refine it by also requiring
  that attribute `attr` is equal to `val`."
  [query-struct attr val]
  (match [query-struct]
         ;; For reasons I don't fully understand, the chef server appears
         ;; to treat *:* as a special case, being totally unwilling
         ;; to allow it to be conjoined with other search terms as
         ;; in e.g. "*:* AND name:foo". However "name:* AND name:foo"
         ;; and "*:foo AND name:foo" are fine.
         [[{:attr "*" :val "*" :type "FIELD"}]]
         [{:attr attr :val val :type "FIELD"}]
         ;; Otherwise tack on "AND attr:field" to the query, if it's
         ;; not already there.
         [_]
         (let [conjunct {:attr attr :val val :type "FIELD" :conjoiner "AND"}]
           (if (some #(= % conjunct)
                     (assoc-in query-struct [0 :conjoiner] "AND"))
             query-struct
             (conj query-struct conjunct)))))
