(ns reagent-manage.search-interface
  (:require
   [reagent.session :as session]
   [reagent-manage.utils :as u]
   [reagent-manage.popup :as popup]
   [reagent-manage.query-builder :as qb]
   [reagent-manage.api :as api]
   [reagent-manage.panels :as panels]
   [reagent.core :as reagent :refer [atom cursor]]
   [reagent-manage.ajax :refer [GET POST PUT DELETE]]
   [cljs.core.match :refer-macros [match]])
  (:require-macros
   [reagent-manage.macros :refer [<? go-api]]))

(def initial-search
  "Default search data structure."
  {:index "node"
   :id ""
   :query ""
   :description ""
   })

(defn initial-search-index
  "Default search data structure with a specified index."
  [index]
  (merge initial-search {:index index
                         :raw-query "*:*"
                         :query-struct (qb/parse-string "*:*")}))

(defn set-hash!
  "Unconditionally set URI fragment."
  ([v] (set-hash! (.-location js/window) v))
  ([loc v] (aset loc "hash" v)))

(defn set-hash-if-different!
  "Set URI fragment only if it is different from the current one."
  ([v] ;; v includes "#" character already
   (if (not= (.-state js/history) v) (js/history.pushState v v v))))

(def search-page-size
  "How many search results to load at once."
  200)

(defn merge-results
  "Combine previously loaded search results with a newly loaded page."
  [x y]
  (match [x y]
         [nil y] y
         [x nil] x
         [_ _] (merge x {:rows (vec (concat (:rows x) (:rows y)))})))

(defn execute-search
  "Execute a search on the Chef Infra Server for the search described in
  atom `doc-atm`. `start` and `rows` have the same meaning as in the
  Chef Infra Server API."
  ([doc-atm] (execute-search doc-atm 0))
  ([doc-atm start] (execute-search doc-atm start search-page-size))
  ([doc-atm start rows]
   (popup/clear)
   (let [doc @doc-atm
         index (:index doc)]
     (go-api
      (let [results (<? (api/request-absolute
                         GET
                         (api/convert-to-api-uri (get (api/search-indexes) index))
                         {:params {"q" (qb/get-doc-query doc)
                                   "rows" rows
                                   "start" start}}))]
        (let [old-results (get-in @doc-atm [:search-results :results] nil)
              new-results (if (= 0 start)
                            results
                            (merge-results old-results results))]
          (swap! doc-atm assoc
                 :search-results
                 {:index (keyword index)
                  :results new-results})))))))

(defn set-current-search!
  "Set the current thing being viewed to the supplied saved search,
  `:new` if viewing a non-saved search."
  [satm search-id]
  (reset! (cursor satm [:current-entity]) [:saved-search search-id]))

(defn new-search-with-index
  "Replace the current search a fresh search for all items in index `index`."
  [index satm]
  (let [new-search-atm (cursor satm [:search-data :new])]
    (set-current-search! satm :new)
    (reset! new-search-atm
            (initial-search-index index))
    (execute-search new-search-atm)))

(defn externalize-doc
  "Convert search interface state into a saved-search record."
  [doc]
  {:id (doc :id)
   :index (doc :index)
   :query (qb/get-doc-query doc)
   :description (doc :description)})

(defn internalize-doc
  "Convert a saved-search record into a search interface state."
  [doc]
  (merge doc {:raw-query (:query doc)
              :query-struct (qb/parse-string (:query doc))
              :orig doc}))

(defn get-saved-search-into-atom
  "Put the saved search at uri `uri` into atom `atom`."
  [uri atom]
  (go-api
   (let [doc (internalize-doc (<! (api/request-absolute GET (api/convert-to-api-uri uri))))]
     (reset! atom doc)
     (execute-search atom))))

(defn load-saved-searches
  "Load the list of saved searches, into the appropriate place in
  `satm`, the global state atom."
  [satm]
  ;; Even though we have an explicit try block below to catch the
  ;; possibility of a 404 when getting the saved-searches data-bag, we
  ;; still want to display any *other* exceptions in the usual way, so
  ;; we still call (go-api) here rather than a naked (go).
  (go-api
   (try
     (let [;; we need the api listing for the url it gives us...
           saved-searches-list (<? (api/request GET "/data/saved-searches"))
           ;; and also the search results are a convenient way of finding out the
           ;; data of every saved search without doing a bunch of separate GET requests.
           saved-searches-search (<? (api/request GET "/search/saved-searches"))]
       (swap! satm assoc :saved-searches saved-searches-list)
       (swap! (cursor satm [:search-data])
              #(merge %
                      (u/make-map (for [row (:rows saved-searches-search)]
                                    [(:id row) (internalize-doc row)])))))
     (catch js/Object e
       (when (not= (:status e) 404) (throw e))
       ;; If we've gotten here, we *did* have a 404, so just try
       ;; create the saved-search data-bag.
       (<? (api/request POST "/data" {:params {:name "saved-searches"}}))
       (swap! satm assoc :saved-searches [])))))

(def schemas
  "Information about the columns of known search indices."
  {:node
   [[:name "Name" :link]
    [:platform "Platform"]
    [:fqdn "FQDN"]
    [:chef_environment "Environment"]]
   :environment
   [[:name "Name" :link]
    [:description "Description"]]
   :client
   [[:name "Name" :link]]
   :role
   [[:name "Name" :link]]
   :saved-searches
   [[:id "Name" :link]
    [:description "Description"]
    [:query "Query"]
    [:index "Index"]]})

(defn get-schema-for-index
  "Get a schema (i.e. a list of pairs of column keys and
  human-readable names) for the given index."
  [index]
  (or (get schemas index) [[:id "Name" :link]]))

(defn results-column-header
  "Component to render a column header in search results."
  [{:keys [col-id cname results-left doc-atm]}]
  (let [sort-order-atm (cursor doc-atm [:search-results :sort-order])
        remaining-results? (not= results-left 0)
        popup-name (if remaining-results? col-id nil)]
   [:th (merge
         {:className (match @sort-order-atm
                            [col-id :asc] "sortField"
                            [col-id :desc] "sortField reverse"
                            _ [:span])
          :on-mouse-down
          (fn [e]
            (if remaining-results?
              (popup/toggle col-id)
              (swap! sort-order-atm
                     #(match [results-left %]
                             [0 [col-id :asc]] [col-id :desc]
                             [0 _] [col-id :asc]
                             [_ _] [col-id :invalid])))
            (.preventDefault e))}
         (if remaining-results?
           {:style {:cursor "not-allowed"}}
           {}))

    [popup/popup-on [:span cname] popup-name
     [:span {:on-mouse-down #(.stopPropagation %)}
      "Columns can be sorted when all results are loaded." [:br]
      [:a {:on-click #(execute-search
                       doc-atm
                       (count (get-in @doc-atm [:search-results :results :rows]))
                       results-left)}
       "Load " results-left " more results."]]]]))

(defn compare-col-asc
  "Sort on column `col-id`, ascending."
  [col-id]
  (u/comparef #(get % col-id)))

(defn compare-col-desc
  "Sort on column `col-id`, descending."
  [col-id]
  (fn [a b] (- ((u/comparef #(get % col-id)) a b))))

(defn sort-rows
  "Sort rows `rows` using the sort order `sort-order`."
  [rows sort-order]
  (match sort-order
         [col-id :asc] (sort (compare-col-asc col-id) rows)
         [col-id :desc] (sort (compare-col-desc col-id) rows)
         _ rows))

(defn refine-search!
  "Take the search in `doc-atm` and refine it to also require field
  `attr` is `val`."
  [doc-atm attr val]
  (swap!
   (cursor doc-atm [:query-struct])
   #(qb/refine-search % attr val))
  (reset! (cursor doc-atm [:raw-query]) nil))

(defn search-results-link
  "Compute one search result table cell's link."
  [{:keys [value link-url doc-atm index] :as cell-info}]
  (let [real-url (let [spacer (if (re-find #"data_bags" link-url) "items/" "")]
                   (str link-url spacer value))]
    (if (= index :node)
      [:a {:on-click #(reset! (cursor session/state [:current-entity]) [:node real-url])} value]
      [:a {:href real-url} value])))

(defn search-results-cell
  "Component to render one search result table cell."
  [{:keys [coords value link? doc-atm] :as cell-info}]
  (if link?
    [:td (search-results-link cell-info)]
    [:td {:class "refinable"
          :on-mouse-down
          #(do (popup/toggle coords) (.preventDefault %))}
     [popup/popup-on [:span value] coords
      [:a {:on-mouse-down
           #(do
              (refine-search! doc-atm (name (:column coords)) value)
              (execute-search doc-atm)
              (.stopPropagation %))}
       "Refine search by this field"]]]))

(defn search-results-row
  "Component to render one search result row. The `row-number` is required
  only for having a unique set of coordinates to place search-refining popups.
  `index` is the search index (e.g. node, client, etc.) which is used to create
  the appropriate url for making links to entity views."
  [{:keys [row-number schema result index doc-atm]}]
  (let
      [index-name (case index
                    (:node :role :environment :client) (str (name index) "s")
                    (str "data_bags/" (name index)))
       link-url (when index (str (api/base-url) "/" index-name "/"))]
    [:tr (for [[column _ & opt] schema]
           ^{:key column} [search-results-cell
                           {:coords {:row-number row-number :column column}
                            :value (get result column)
                            :link? opt
                            :link-url link-url
                            :index index
                            :doc-atm doc-atm}])]))

(defn search-results
  "Component to render all search results."
  [doc-atm search-data]
  (if (nil? search-data) nil
      (let [{:keys [results index]} search-data
            schema (get-schema-for-index index)
            sort-order-atm (cursor doc-atm [:search-results :sort-order])
            results-left (- (:total results) (count (:rows results)))]
        (if (= 0 (:total results))
          [:h3 "No Results"]
          [:div
           [:table.search-results
            [:thead
             [:tr (for [[column cname] schema]
                    ^{:key column} [results-column-header
                                    {:col-id column
                                     :cname cname
                                     :results-left results-left
                                     :doc-atm doc-atm}])]]
            `[~:tbody
              ~@(for [[n result] (u/indexify (sort-rows (:rows results) @sort-order-atm))]
                  (do
                    ^{:key (get result (first (first schema)))}
                    [search-results-row {:schema schema
                                         :result result
                                         :index index
                                         :row-number n
                                         :doc-atm doc-atm}]))
              ~@(if (not= 0 results-left)
                  [[:tr [:td
                         {:style {:text-align "center"} :col-span (count schema)}
                         [:a {:on-click
                              #(execute-search doc-atm (count (:rows results)))}
                          [:i.fi-arrow-down] " " results-left " More Results"]]]]
                  [])]]]))))

(defn form-row
  "Component to render one row of the saved-search metadata form."
  [label input]
  [:div.row.collapse
   [:div.medium-2.columns [:label.inline.right label]]
   [:div.medium-8.columns  input]
   [:div.medium-2.columns]])

(defn changed-name?
  "Returns true if the saved-search name in `doc` is different
  from (i.e. has been edited by the user) compared to the more global notion
  of which saved-search we are currently editing."
  [satm doc]
  (match [@(cursor satm [:current-entity])]
         [[:saved-search current-search]]  (not= current-search (:id doc))
         [_]                               true))

(defn submit-button
  "Button to save new or existing saved search."
  [satm doc method]
  (let [external-doc (externalize-doc doc)]
   [:button.button.tiny.success
    {:disabled (and (= method "PUT") (= (:orig doc) external-doc))
     :on-click (fn [e]
                 (let [base "/data/saved-searches"]
                   (.preventDefault e)
                   (go-api
                     (<? (api/request
                          (case method
                            "PUT" PUT
                            "POST" POST)
                          (case method
                            "PUT" (str base "/" (doc :id))
                            "POST" base)
                          {:params external-doc}))
                     (when (= method "POST")
                       (reset! (cursor satm [:search-data (doc :id)]) doc)
                       (reset! (cursor satm [:search-data :new])
                               initial-search)
                       (set-current-search! satm (doc :id))
                       (load-saved-searches satm))
                     (do (reset! (cursor satm [:search-data (doc :id) :orig])
                                 external-doc)))))}
    [:i.fi-check] " Save"]))

(defn delete-button
  "Button to delete saved search."
  [satm doc]
  [:button.button.tiny.alert {:on-click (fn [e]
                        (.preventDefault e)
                        (go-api
                          (<? (api/request
                               DELETE
                               (str "/data/saved-searches/" (doc :id))))
                          (load-saved-searches satm)
                          (set-current-search! satm :new)
                          (reset! (cursor satm [:search-data :new])
                                  @(cursor satm [:search-data (doc :id)]))))}
   [:i.fi-x] " Delete"])

(defn execute-button
  "Button to execute current search."
  []
  [:button.primary.postfix {:type "submit"}
   [:i.fi-magnifying-glass] " Search"])

(defn advanced-search-button [satm]
  "Button to toggle advanced search panel."
  [:a.panel-opener
   {:on-click #(panels/toggle satm :advanced-search)}
    "Advanced Search"])

(defn saved-search-button [satm]
  "Button to toggle saved search panel."
  [:a.panel-opener
   {:on-click #(panels/toggle satm :saved-search)}
    "Saved Search"])

(defn form-buttons
  "Collection of buttons for saved search."
  [satm doc method]
  `[:div.row.collapse [:span.right
      ~[submit-button satm doc method] " "
      ~@(case method
          "PUT" [" " [delete-button satm doc]]
          "POST" [])]])

(defn form
  "Component to render the search interface."
  [satm method doc-atm]
  (if-let [doc @doc-atm]
    (let [show-advanced-search (panels/visible? satm :advanced-search)
          show-saved-search (panels/visible? satm :saved-search)

          ;; @query-struct is always a data structure that represents a
          ;; search.
          query-struct (cursor doc-atm [:query-struct])

          ;; @raw-query may be nil, in which case it means we should
          ;; currently trust @query-struct to be an acurate
          ;; representation of the query, or else some string, which we
          ;; should take to be the user's textual modification of the query string.
          raw-query (cursor doc-atm [:raw-query])]
      [:div
        (if (panels/visible? satm :welcome-message)
          [:div.data-alert.alert-box.radius.info
            [:span
              [:p "This is our new search feature. It's experimental, so you might find a few bugs. You can return to Chef Manage at any time using the " [:a {:href "/"} "Manage"] " link in the upper-right corner."]
              [:a.close {:on-click #(panels/perma-hide satm :welcome-message)} u/times ]]
            [:span]])
       [:form.search-form
        {:on-submit #(do (.preventDefault %)
                         (execute-search doc-atm))}
        [:div.row.collapse
        [:div.medium-2.large-2.columns
         `[~u/select-input ~(cursor doc-atm [:index]) {}
           ~@(for [index (keys (api/search-indexes))]
               (u/select-option index index))]]
        [:div.medium-8.large-8.columns
         [qb/search-query-display
          (cursor doc-atm [:query-struct])
          (cursor doc-atm [:raw-query])]]
        [:div.medium-2.large-2.columns
         [execute-button]]]]
       (if show-advanced-search
         [:div.row.collapse
          [:div.medium-12.large-12.columns
           [qb/search-builder-component
            (qb/write-listener query-struct (fn [] (reset! raw-query nil)))
            #(panels/hide satm :advanced-search)]]]
         [:div.row.collapse
          [:div.medium-10.large-10.columns]
          [:div.medium-2.large-2.columns
           [advanced-search-button satm]]])
       (if show-saved-search
         [:div.closable-container
          [form-row "Name" [u/text-input (cursor doc-atm [:id])]]
          [form-row "Description" [u/text-input (cursor doc-atm [:description])]]
          [form-buttons satm doc (if (changed-name? satm doc) "POST" method)]
          [:a.close {:on-click #(panels/hide satm :saved-search)} u/times]]
         [:div.row.collapse
          [:div.medium-10.large-10.columns]
          [:div.medium-2.large-2.columns
           [saved-search-button satm]]])
       [search-results doc-atm (:search-results doc)]])
    [:span]))
