(ns reagent-manage.utils
  (:require [reagent.core :as reagent :refer [atom cursor]]))


;; Views & Utilities
;; -------------------------
;; A couple of simple wrappers around html elements. These sorts of
;; things might be in reagent-forms, I haven't looked yet.
;;
(defn text-input
  ([value-atom] (text-input value-atom {}))
  ([value-atom opts]
   [:input (merge {:type "text"
                   :value @value-atom
                   :on-change #(reset! value-atom (-> % .-target .-value))} opts)]))

(defn textarea-input
  ([value-atom] (textarea-input value-atom {}))
  ([value-atom opts]
   [:textarea
    (merge {:value @value-atom
            :style {:resize "none"}
            :on-change #(reset! value-atom (-> % .-target .-value))} opts)]))

(defn select-input [value-atom attrs & options]
  [:select (merge {:value @value-atom
                   :on-change #(reset! value-atom (-> % .-target .-value))}
                  attrs)
   options])

(defn select-option [value text]
  ^{:key value} [:option {:value value} text])

(defn without-nth
  "Sequence `seq` as a vector with its `n`th element removed."
  [seq n]
  vec (concat (subvec seq 0 n) (subvec seq (+ 1 n))))

(defn wrap-reader
  "Wrap `ratom` with a function `reader` that intercedes when
  you try to read its value out. Writes to it work as normal."
  [ratom reader]
  (reagent.core/wrap (reader @ratom) reset! ratom))

(defn indexify
  "Given a sequence `xs`, return a sequence consisting of pairs `[i
  x]` where `i` is the index in the sequence and `x` is an element of
  the sequence. Useful for wrapping around a thing that you're already
  doing a `for` over when you need to know the index of each item."
  [xs]
  (let [vxs (vec xs)]
    (for [ix (range (count vxs))]
      [ix (get vxs ix)])))

(defn error?
  "Return true iff the argument is an Error."
  [x]
  (instance? js/Error x))

(defn make-map
  "Make a map from a sequence of pairs"
  [seq]
  (apply hash-map (apply concat seq)))

(defn make-err
  "Make an error object carrying first-class data. We set the field
  \"payload\" of the Error object to the error data, because the Error
  constructor argument is stringified. We still supply the constructor
  argument just in case the human-readable string is useful to anybody
  downstream."
  [e]
  (let [err (js/Error. e)]
    (aset err "payload" e)
    err))

(defn throw-err
  "If the argument is an Error, throw its payload."
  [x]
  (if (error? x)
    (throw (.-payload x))
    x))

(defn comparef
  "Derive sort comparison function from a measure function `f`."
  [f]
  (fn [a b] (compare (f a) (f b))))

;; Unicode times symbol"
(def times "×")

(defn set-toggle
  "Returns set `s` with `item` removed or added, whichever causes the
  result to differ from the original set."
  [s item]
  ((if (contains? s item) disj conj) s item))

(defn get-config
  "Get configuration option `option` from rails-supplied configuration data."
  [option]
  (aget (.-config js/Chef) option))
