(ns reagent-manage.query-builder-test
  (:require
   [cljs.test :refer-macros [deftest testing is]]
   [cljs.core.match :refer-macros [match]]
   [reagent.core :as reagent :refer [atom cursor]]
   [reagent-manage.query-builder :as query]))

(deftest base
  (testing "update-in"
    (let [example-map {:a 0 :b {:c {}}}]
      (is (= (query/update-in example-map [] #(assoc % :d 3))
             {:a 0 :b {:c {}} :d 3}))
      (is (= (query/update-in example-map [:b :c] #(assoc % :d 3))
             {:a 0 :b {:c {:d 3}}}))))

  (testing "write listener"
    (let [atm-1 (atom 0)
          atm-2 (atom 0)
          wr-atm (query/write-listener atm-1 #(swap! atm-2 inc))]
      (swap! atm-1 inc)
      (is (= @atm-1 1))
      (is (= @atm-2 0))
      (swap! wr-atm inc)
      (is (= @atm-1 2))
      (is (= @atm-2 1))))

  (testing "simple query parsing"
    (is (= (query/parse-string "*:") nil))
    (is (= (query/parse-string "*:*") [{:attr "*", :val "*", :type "FIELD"}])))

  (testing "range query parsing"
    (is (= (query/parse-string "value:[1 TO 9]")
           [{:type "INC_RANGE", :from "1", :upto "9", :attr "value"}]))
    (is (= (query/parse-string "value:{1 TO 9}")
           [{:type "EXC_RANGE", :from "1", :upto "9", :attr "value"}])))

  (testing "query round trip"
    (let [orig-query "a:b AND c:d"
          manual-string "some other string"
          parsed (query/parse-string orig-query)]
      (is (= parsed
             [{:attr "a" :val "b" :type "FIELD"} {:attr "c" :val "d" :type "FIELD" :conjoiner "AND"}]))
      (is (= (query/resolve-query-struct parsed nil)
             [orig-query true]))
      (is (= (query/resolve-query-struct parsed orig-query)
             [orig-query true]))
      (is (= (query/resolve-query-struct parsed manual-string)
             [manual-string false]))))

  (testing "query parsing whitespace"
    (is (= (query/parse-string "") []))
    (is (= (query/parse-string "   ") []))
    (is (= (query/parse-string "   foo:bar  OR baz:mumble   ")
           [{:attr "foo", :val "bar", :type "FIELD"}
            {:attr "baz", :val "mumble", :type "FIELD" :conjoiner "OR"}])))

  (testing "refine-search"
    (is (= (query/refine-search
            (query/parse-string "*:*") "platform" "centos")
           (query/parse-string "platform:centos")))
    (is (= (query/refine-search
            (query/parse-string "foo:bar") "platform" "centos")
           (query/parse-string "foo:bar AND platform:centos")))
    (is (= (query/refine-search
            (query/parse-string "platform:centos") "platform" "centos")
           (query/parse-string "platform:centos")))
    (is (= (query/refine-search
            (query/parse-string "foo:bar AND platform:centos") "platform" "centos")
           (query/parse-string "foo:bar AND platform:centos")))
    (is (= (query/refine-search
            (query/parse-string "platform:centos AND foo:bar") "platform" "centos")
           (query/parse-string "platform:centos AND foo:bar"))))
  )
