Chef::Manage::Application.routes.draw do
  root :to => 'dashboard#show'
  get '_status' => 'application_status#show'
  get 'changelog' => 'changelog#show'

  # Login/out routes
  get 'login' => 'sessions#new', :as => 'login'
  post 'login' => 'sessions#create'
  post 'logout' => 'sessions#destroy', :as => 'logout'

  # OIDC routes
  get 'oidc' => 'oidc#new', :as => 'oidc_login'
  get 'oidc/callback' => 'oidc#callback', :as => 'oidc_callback'

  # Sign Up routes
  get 'signup' => 'accounts#new', :as => 'signup'
  get 'verify_email' => 'users#verify_email', :as => 'verify_email'

  if Settings.marketplace.sign_up.enabled
    post 'signup' => 'accounts#create'
  else
    post 'signup' => 'accounts#send_verification'
  end

  post 'signup/psc' => 'accounts#create_psc'

  get 'signup/validity/:type' => 'accounts#validate_signup'
  get 'signup/verify' => 'accounts#verify'
  post 'signup/verify' => 'accounts#create'

  # account linking
  resource :account_links, path: '/link-account', only: [:show, :create, :update]

  # license
  get 'license' => 'license#show'

  # starter kit
  get 'starter-kit' => "organizations#get_starter_kit"

  # Password reset
  resource :password_reset, path: '/password-reset', except: :destroy

  # System recovery
  resource :system_recovery, path: '/system-recovery', only: [:create, :show]

  # Main application routes
  constraints(:id => /[^\/]+/) do
    resources :organizations do
      member do
        get 'exists'
        get 'getting_started'
        get 'knife_config'
        post 'getting_started' => 'organizations#starter_kit'
        post 'leave'
      end
      get 'details', on: :member, action: 'show'
      get 'members', on: :member, action: 'show'

      resources :actions

      resources :association_requests,
                :controller => 'organization_association_requests',
                :only => :create

      resources :nodes do
        get 'run_lists', on: :member
        get 'details', on: :member, action: 'show'
        get 'attributes', on: :member, action: 'show'
        get 'permissions', on: :member, action: 'show'
      end
      resources :roles do
        get 'run_lists', on: :member
        get 'details', on: :member, action: 'show'
        get 'attributes', on: :member, action: 'show'
        get 'permissions', on: :member, action: 'show'
      end
      resources :data_bags do
        resources :items, :controller => :data_bag_items do
          get 'details', on: :member, action: 'show'
        end
        get 'permissions', on: :member, action: 'show'
      end
      resources :runs do
        get 'details', on: :member, action: 'show'
        get 'error_log', on: :member, action: 'show'
        get 'run_list', on: :member, action: 'show'
      end
      get "runs_query" => "reporting_query#query"
      # Dashboard routes
      get 'dashboard' => 'reporting_dashboard#show'
      resources :cookbooks do
        member do
          get 'fetchfile'
          get ':version_id(/:tab)', action: 'show_version',
                                      constraints: { version_id: /[\w\.]+/ }
        end
      end
      resources :users do
        get 'groups', on: :member
        get 'details', on: :member, action: 'show'
        member do
          post 'remove'
        end
        put 'password', on: :member, action: 'update_password'
      end
      resources :groups do
        get 'details', on: :member, action: 'show'
        get 'members', on: :member, action: 'show'
        get 'group_permissions', on: :member, action: 'show'
      end
      resources :clients do
        get 'details', on: :member, action: 'show'
        get 'permissions', on: :member, action: 'show'
      end
      resources :acls, constraints: { id: /.*/ }
      resources :environments do
        get 'recipes', on: :member
        get 'cookbooks', on: :member
        get 'details', on: :member, action: 'show'
        get 'attributes', on: :member, action: 'show'
        get 'permissions', on: :member, action: 'show'
      end
      get 'global_permissions(/:container(/permissions))' => 'dashboard#show'

      resources :search, :only => [:index, :show]
    end
  end

  # user invitations
  resources :users, :only => [] do
    resources :association_requests,
              :controller => 'user_association_requests',
              :only => [:index, :update]
  end

  get 'api/organizations/:organization_id/*path' => 'chef_server_api#show',
        constraints: lambda { |request| request.format == 'json' }

  # 404 catcher, MUST BE THE LAST ROUTE
  unless Rails.application.config.consider_all_requests_local
    match '*not_found' => 'errors#error_404', :via => :all
  end
end
