require 'time'

class ChefManageUserStore

  def initialize(url)
    @redis = Redis.new(:url => url)
  end

  def get_user_record(user)
    record = @redis.get(user)

    if record.nil?
      Hash.new
    else
      JSON.parse(record)
    end
  end

  # Initializes password reset for new users/sessions
  def initialize_password_reset(user)
    password_reset(user)
  end

  # Resets the last password reset time for the supplied user to current time and
  # sets the password reset signature for validation; signature is optional (only
  # stored when a password reset link is generated)
  def password_reset(user, signature = "")
    user_record = get_user_record(user)

    # Don't use symbols for hash keys here, JSON will turn them into strings
    user_record["password_reset_time"] = Time.zone.now
    # We just want the actual signature; if we leave this to to_json, we'll get a
    # hash-encoded object, to_s gives us the actual token
    user_record["password_reset_signature"] = signature.to_s
    @redis.set(user, user_record.to_json)
  end

  # Erases the signature when the password is reset
  def invalidate_password_reset_signature(user)
    user_record = get_user_record(user)
    user_record["password_reset_signature"] = ""
    @redis.set(user, user_record.to_json)
  end

  # Check the password reset signature
  def valid_password_reset_signature(user, signature)
    record = @redis.get(user)

    if record
      user_record = JSON.parse(record)

      if user_record["password_reset_signature"] == signature
        return true
      end
    end

    false
  end

  # This returns true if the time supplied is more recent than the last password
  # reset time for the supplied user
  def check_password_reset_time(user, time)
    record = @redis.get(user)

    if record
      user_record = JSON.parse(record)

      # Time is stored as a string, so we'll need to parse it here to get the actual
      # timestamp value
      if user_record["password_reset_time"] &&
          Time.parse(user_record["password_reset_time"]) > time
        return true
      end
    end

    false
  end

  # Login failure functions; one for incrementing it on failed logins, one for
  # resetting it on success, and one for getting number of failures
  def increment_login_failures(user)
    user_record = get_user_record(user)

    if user_record["login_failures"]
      user_record["login_failures"] += 1
    else
      user_record["login_failures"] = 1
    end

    @redis.set(user, user_record.to_json)
  end

  def reset_login_failures(user)
    user_record = get_user_record(user)
    user_record['login_failures'] = 0
    @redis.set(user, user_record.to_json)
  end

  def login_failure_count(user)
    record = @redis.get(user)

    if record
      user_record = JSON.parse(record)

      if user_record["login_failures"]
        return user_record["login_failures"]
      end
    end

    0
  end
end
