# Helpers for setting up a Chef::ServerAPI connection
require 'chef'
require 'veil'

module ChefRestHelpers

  include SecretsHelpers
  Chef::ServerAPI::HTTPRequest.user_agent =  "Chef Manage#{Chef::ServerAPI::HTTPRequest::UA_COMMON}"

  # Takes a Net::HTTPServerException, usually returned from the Chef server,
  # and attempts to parse the body as JSON. If that doesn't work, it attempts
  # to give a more generic error
  def parsed_json_for_error(error)
    sanitizer = Rails::Html::FullSanitizer.new
    JSON.parse(sanitizer.sanitize(error.response.body))
  rescue JSON::ParserError
    { 'error' => error.message }
  end

  protected

  def server(user, organization)
    # raise "Username and organization are required" unless user.present? && organization.present?
    _server(user, "#{platform_config.url}/organizations/#{organization}")
  end

  def server_root
    @server_root ||= _server(platform_config.user, platform_config.url)
  end

  # User the root url as a specified user
  def server_user_root(user)
    _server(user, platform_config.url)
  end

  def _server(user, url)
    raise ArgumentError.new "A privileged key is not available on this system. Ensure that you have run 'chef-server-ctl reconfigure' before continuing." unless defined? KEY

    # This is configurable using /etc/chef-manage/manage.rb file on Chef Infra Server. The default ssl_verify_mode is :verify_none
    # The defaults are set in `omnibus/files/cookbooks/omnibus-chef-manage/attributes/default.rb` but in case that config file isn't propertly recreated, we fall back again here.
    # in recipes/config.rb we convert the config hash to JSON and back before YAML. This removes any symbols which YAML otherwise seems to support, so we have to force back to a symbol here.
    Chef::Config[:ssl_verify_mode] = Settings.ssl_verify_mode.nil? ? :verify_none : Settings.ssl_verify_mode.to_sym
    Chef::Config[:trusted_certs_dir] = Settings.trusted_certs_dir || ""
    unless Chef::Config[:ssl_verify_mode] == :verify_none
      Chef::Config[:ssl_client_cert] = Settings.ssl_client_cert
      Chef::Config[:ssl_client_key] = Settings.ssl_client_key
      Chef::Config[:ssl_ca_file] = Settings.ssl_ca_file
    else
      Chef::Config.delete(:ssl_client_cert)
      Chef::Config.delete(:ssl_client_key)
      Chef::Config.delete(:ssl_ca_file)
    end

    # NOTE: setting api version to 0 as some existing functionalities related to setting keys for user not working with api V1
    Chef::ServerAPI.new(
      url,
      client_name: user,
      signing_key_filename: nil,
      headers: chef_rest_helper_headers,
      raw_key: KEY,
      api_version: "0"
    )
  end

  private

  def chef_rest_helper_headers
    @chef_rest_helper_headers ||= { 'X-Ops-Request-Source' => 'web' }.merge(
      platform_config.webkey_tag ? { 'X-Ops-Webkey-Tag' => platform_config.webkey_tag }
                                 : {})
  end

  def platform_config
    @platform_config ||= Settings.platform
  end
end
