require 'openid_connect'

class OidcClient

  attr_reader :client_id, :client_secret, :redirect_uri, :issuer_url,
    :verify_tls, :id_token, :client

  def initialize(client_id, client_secret, redirect_uri, issuer_url, verify_tls = true)
    @client_id = client_id
    @client_secret = client_secret
    @redirect_uri = redirect_uri
    @issuer_url = issuer_url
    @verify_tls = verify_tls
    @client ||= OpenIDConnect::Client.new client_config
  end

  def ready?
    !provider_config.nil?
  end

  def authorization_uri(opts = {})
    return unless ready?
    client.authorization_uri(opts)
  end

  def pub_key
    provider_config.public_keys.first
  end

  def provider_config
    # TODO: intelligent caching? pubkey may change
    @provider_config ||= discover_provider_config
  end

  def client_config
    config = {
      identifier: client_id,
      secret: client_secret,
      redirect_uri: redirect_uri
    }

    # work-around, OpenIDConnect::Discovery::Provider::Config.discover! hates
    # http -- anyways, if it's http, it's probably a delivery dev environment
    if issuer_url.starts_with? 'http://'
      config[:authorization_endpoint] = issuer_url + '/oidc/auth'
      config[:token_endpoint] = issuer_url + '/oidc/token'
    elsif ready?
      config[:authorization_endpoint] = provider_config.authorization_endpoint
      config[:token_endpoint] = provider_config.token_endpoint
    end

    config
  end

  def exchange_code_for_token(code)
    client.authorization_code = code
    begin
      access_token = client.access_token! :client_auth_body
      @id_token = OpenIDConnect::ResponseObject::IdToken.decode access_token.id_token, pub_key
    rescue Rack::OAuth2::Client::Error => e
      Rails.logger.error("Failed OIDC code exchange: #{e.response}")
      nil
    end
  end

  def user_from_token(token)
    {
      'username' => namify(token.sub),
      'display_name' => token.sub,
      'email' => token.raw_attributes['email'],
      'first_name' => token.raw_attributes['first'],
      'last_name' => token.raw_attributes['last'],
      'external_authentication_uid' => token.sub,
      'recovery_authentication_enabled' => false,
    }
  end

  private

  # if this is unset or true, don't disable
  def disable_tls?
    verify_tls == false
  end

  def disable_tls
    OpenIDConnect.http_config do |config|
      config.ssl_config.verify_mode = OpenSSL::SSL::VERIFY_NONE
    end
  end

  def discover_provider_config
    # first contact with issuer is here
    disable_tls if disable_tls?

    begin
      OpenIDConnect::Discovery::Provider::Config.discover! issuer_url
    rescue OpenIDConnect::Discovery::DiscoveryFailed => e
      Rails.logger.error("Failed OIDC provider discovery: #{e.inspect}")
      nil
    end
  end

  def namify(uid)
    # Chef Server defines USERNAME_REGEX as "[a-z0-9\-_]+").
    uid.gsub(/@.*$/, '') # drop host part if it's an email address
       .downcase
       .delete('^a-z0-9\-_') # drop any unsupported chars
  end
end
