require 'zip'
require 'config_rb'

# Represents the starter kit zip file
class StarterKit
  include ApplicationHelper
  attr_reader :orgname, :user, :username

  # Takes an org name and a user hash
  def initialize(organization_name, user)
    @orgname = organization_name
    @user = user
    @username = user['username']
  end

  # Content disposition header
  def content_disposition
    "#{disposition}; filename=\"#{filename}\""
  end

  # Hash of options to be used in a controller's send_data method
  def send_data_options
    { type: 'application/zip',
      disposition: disposition,
      filename: filename }
  end

  # Serializes the starter kit as a zip stream
  def to_zip
    add_static_content_to_zip!
    add_dynamic_directories_to_zip!
    add_dynamic_content_to_zip!
    zip.close_buffer.string
  end

  private

  def add_dynamic_content_to_zip!
    zip.put_next_entry('chef-repo/.chef/config.rb')
    zip.write(config_rb.to_s)
    zip.put_next_entry('chef-repo/roles/starter.rb')
    zip.write(render('roles/starter.rb'))
    zip.put_next_entry("chef-repo/.chef/#{username}.pem")
    zip.write(user['private_key'])
  end

  def add_dynamic_directories_to_zip!
    dynamic_directories.each do |dir|
      zip.put_next_entry("chef-repo/#{dir}/")
    end
  end

  def add_static_content_to_zip!
    Dir[File.join(source_path, '**', '{*,.*}')].each do |file|
      base_path = file.sub(File.join(source_path, ''), '')
      if File.directory?(file)
        zip.put_next_entry("chef-repo/#{base_path}/")
      else
        zip.put_next_entry("chef-repo/#{base_path}")
        zip.write(File.new(file, 'rb').read)
      end
    end
  end

  def dynamic_directories
    %w[ .chef roles ]
  end

  # Disposition type used in header and send data options
  def disposition
    'attachment'
  end

  def filename
    'chef-starter.zip'
  end

  def config_rb
    @config_rb ||= ConfigRb.new(orgname, username)
  end

  def render(filename)
    ERB.new(open(
      "#{ApplicationController.view_paths.first}/organizations/starter_kit/#{filename}.erb"
    ).read).result(binding)
  end

  def source_path
    @source_path ||= Rails.root.join('vendor', 'starter_kit')
  end

  def user_display_name
    "#{user['first_name']} #{user['last_name']}"
  end

  def zip
    @zip ||= Zip::OutputStream.new(StringIO.new, true)
  end
end
