require 'rest_client'
require 'uri'
require 'mixlib/authentication'
require 'chef/sandbox' # So we can decode the object returned via JSON >_<

namespace :chef do
  namespace :data do
    def chef_server_rest(args)
      raise "Username is required" unless args[:user]
      raise "Organization is required" unless args[:organization]
      platform_config = Settings.platform
      if args[:user_key]
        headers = {}
      else
        headers = {'x-ops-request-source' => 'web'}
        headers['x-ops-webkey-tag'] = platform_config.webkey_tag if platform_config.webkey_tag
      end
      Chef::ServerAPI.new(
        "#{platform_config.url}/organizations/#{args[:organization]}", client_name: args[:user], signing_key_filename: args[:user_key] || Credentials.get('chef-server', 'webui_key'), :headers => headers)
    end

    desc 'Generate random node data.'
    task :nodes, [:user, :organization, :count] => :environment do |t, args|
      server = chef_server_rest(args)
      factory_args = {:roles => server.get_rest("roles").keys, :recipes => []}
      server.get_rest("cookbooks").each do |cookbook, cookbook_data|
        cookbook_version_data = server.get_rest("cookbooks/#{cookbook}/#{cookbook_data["versions"][0]["version"]}")
        cookbook_version_data.segment_filenames(:recipes).each do |recipe_filename|
          recipe_filename = recipe_filename[0..-4] # Trim the .rb
          factory_args[:recipes] << if recipe_filename == "default"
            cookbook
          else
            "#{cookbook}::#{recipe_filename}"
          end
        end
      end
      args.with_defaults(:count => 100)
      args[:count].to_i.times do
        obj = FactoryBot.build_chef(:node, factory_args)
        raise "Error creating Chef object" unless obj
        puts "Creating #{obj.name}"
        server.post_rest("nodes", obj)
      end
    end

    desc 'Generate random role data.'
    task :roles, [:user, :organization, :count] => :environment do |t, args|
      server = chef_server_rest(args)
      factory_args = {:roles => server.get_rest("roles").keys, :recipes => []}
      server.get_rest("cookbooks").each do |cookbook, cookbook_data|
        cookbook_version_data = server.get_rest("cookbooks/#{cookbook}/#{cookbook_data["versions"][0]["version"]}")
        cookbook_version_data.segment_filenames(:recipes).each do |recipe_filename|
          recipe_filename = recipe_filename[0..-4] # Trim the .rb
          factory_args[:recipes] << if recipe_filename == "default"
            cookbook
          else
            "#{cookbook}::#{recipe_filename}"
          end
        end
      end
      args.with_defaults(:count => 10)
      args[:count].to_i.times do
        obj = FactoryBot.build_chef(:role, factory_args)
        raise "Error creating Chef object" unless obj
        puts "Creating #{obj.name}"
        server.post_rest("roles", obj)
        factory_args[:roles] << obj.name
      end
    end

    desc 'Generate random cookbook data'
    task :cookbooks, [:user, :organization, :user_key, :count] => :environment do |t, args|
      rest = chef_server_rest(args)
      args.with_defaults(:count => 20)
      args[:count].to_i.times do
        obj = FactoryBot.build_chef(:cookbook_version)
        puts "Creating #{obj.name}"
        # Create a sandbox
        new_sandbox = rest.post_rest("sandboxes", { :checksums => obj.checksums })
        # Upload needed files
        by_checksum = obj.manifest_records_by_path.values.inject({}) do |memo, record|
          memo[record[:checksum]] = record
          memo
        end
        new_sandbox['checksums'].each do |checksum, info|
          next unless info['needs_upload']
          url = info['url']
          # Copy-pasta'd from chef/cookbook_uploader
          checksum64 = Base64.encode64([checksum].pack("H*")).strip
          timestamp = Time.now.utc.iso8601
          file_contents = by_checksum[checksum][:content]
          sign_obj = Mixlib::Authentication::SignedHeaderAuth.signing_object(
                                                                             :http_method => :put,
                                                                             :path        => URI.parse(url).path,
                                                                             :body        => file_contents,
                                                                             :timestamp   => timestamp,
                                                                             :user_id     => rest.client_name
                                                                             )
          headers = { 'content-type' => 'application/x-binary', 'content-md5' => checksum64, :accept => 'application/json' }
          headers.merge!(sign_obj.sign(OpenSSL::PKey::RSA.new(rest.signing_key)))
          RestClient::Resource.new(url, :headers=>headers, :timeout=>1800, :open_timeout=>1800).put(file_contents)
        end
        # Finished with the sandbox
        rest.put_rest(new_sandbox['uri'], {:is_completed => true})
        # Upload the cookbook data
        rest.put_rest(obj.save_url, obj)
      end
    end

    desc 'Generate random data bags'
    task :data_bags, [:user, :organization, :count] => :environment do |t, args|
      server = chef_server_rest(args)
      args.with_defaults(:count => 5)
      args[:count].to_i.times do
        obj = FactoryBot.build_chef(:data_bag)
        raise "Error creating Chef object" unless obj
        puts "Creating #{obj.name}"
        server.post_rest("data", obj)
      end
    end

    desc 'Generate random data bag items'
    task :data_bag_items, [:user, :organization, :count, :bag] => :environment do |t, args|
      server = chef_server_rest(args)
      bags = if args[:bag]
        [args[:bag]]
      else
        server.get_rest('data').keys
      end
      args.with_defaults(:count => 20)
      args[:count].to_i.times do
        obj = FactoryBot.build_chef(:data_bag_item, data_bags: bags)
        raise "Error creating Chef object" unless obj
        puts "Creating #{obj.name}"
        server.post_rest("data/#{obj.data_bag}", obj)
      end
    end

  end
end
