require 'test_helper'

describe "Redis::DistributedStore" do
  def setup
    @dmr = Redis::DistributedStore.new [
      { :host => "localhost", :port => "6380", :db => 0 },
      { :host => "localhost", :port => "6381", :db => 0 }
    ]
    @rabbit = OpenStruct.new :name => "bunny"
    @white_rabbit = OpenStruct.new :color => "white"
    @dmr.set "rabbit", @rabbit
  end

  def teardown
    @dmr.ring.nodes.each { |server| server.flushdb }
  end

  it "accepts connection params" do
    dmr = Redis::DistributedStore.new [ :host => "localhost", :port => "6380", :db => "1" ]
    dmr.ring.nodes.size == 1
    mr = dmr.ring.nodes.first
    mr.to_s.must_equal("Redis Client connected to localhost:6380 against DB 1")
  end

  it "forces reconnection" do
    @dmr.nodes.each do |node|
      node.expects(:reconnect)
    end

    @dmr.reconnect
  end

  it "sets an object" do
    @dmr.set "rabbit", @white_rabbit
    @dmr.get("rabbit").must_equal(@white_rabbit)
  end

  it "gets an object" do
    @dmr.get("rabbit").must_equal(@rabbit)
  end

  it "mget" do
    @dmr.set "rabbit2", @white_rabbit
    begin
      @dmr.mget "rabbit", "rabbit2" do |rabbits|
        rabbit, rabbit2 = rabbits
        rabbits.length.must_equal(2)
        rabbit.must_equal(@rabbit)
        rabbit2.must_equal(@white_rabbit)
      end
    rescue Redis::Distributed::CannotDistribute
      # Not supported on redis-rb < 4, and hence Ruby < 2.2.
    end
  end

  it "mapped_mget" do
    @dmr.set "rabbit2", @white_rabbit
    begin
      result = @dmr.mapped_mget("rabbit", "rabbit2")
      result.keys.must_equal %w[ rabbit rabbit2 ]
      result["rabbit"].must_equal @rabbit
      result["rabbit2"].must_equal @white_rabbit
    rescue Redis::Distributed::CannotDistribute
      # Not supported on redis-rb < 4, and hence Ruby < 2.2.
    end
  end

  it "passes through ring replica options" do
    dmr = Redis::DistributedStore.new [
                                    { :host => "localhost", :port => "6380", :db => 0 },
                                    { :host => "localhost", :port => "6381", :db => 0 }
                                ], replicas: 1024
    dmr.ring.replicas.must_equal 1024
  end

  it "uses a custom ring object" do
    my_ring = Redis::HashRing.new
    dmr = Redis::DistributedStore.new [
                                          { :host => "localhost", :port => "6380", :db => 0 },
                                          { :host => "localhost", :port => "6381", :db => 0 }
                                      ], ring: my_ring
    dmr.ring.must_equal my_ring
    dmr.ring.nodes.length.must_equal 2
  end

  describe '#redis_version' do
    it 'returns redis version' do
      @dmr.nodes.first.expects(:redis_version)
      @dmr.redis_version
    end
  end

  describe '#supports_redis_version?' do
    it 'returns redis version' do
      @dmr.nodes.first.expects(:supports_redis_version?).with('2.8.0')
      @dmr.supports_redis_version?('2.8.0')
    end
  end

  describe "namespace" do
    it "uses namespaced key" do
      @dmr = Redis::DistributedStore.new [
        { :host => "localhost", :port => "6380", :db => 0 },
        { :host => "localhost", :port => "6381", :db => 0 }
      ], :namespace => "theplaylist"

      @dmr.expects(:node_for).with("theplaylist:rabbit").returns(@dmr.nodes.first)
      @dmr.get "rabbit"
    end
  end
end
